from ...structure import *

def scoreAcc(M, epochs, X_train, Y_train):
    acc, history = M.gradient_descent(X_train, Y_train, epochs, LearningRateScheduler(LearningRateScheduler.CONSTANT, 0.01, 0.8) , True)
    return acc

def scoreLoss(M, epochs, X_train, Y_train):
    _, history = M.gradient_descent(X_train, Y_train, epochs, LearningRateScheduler(LearningRateScheduler.CONSTANT, 0.01, 0.8) , True)
    return min(1 / history.get_last('loss'), 1.0)


import unittest
import asyncio
import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import numpy as np

# Import your script here, e.g., 
# from myscript import get_reshsper, Reshape, Reshape_forward_prop, Reshape_back_prop, eye_stretch

MAX_ENTRIES = 100  # Match with the script configuration


class TestReshapeFunctions(unittest.TestCase):

    def test_eye_stretch(self):
        a, b = 5, 10
        result = gnn.quaziIdentity.eye_stretch(a, b)
        self.assertEqual(result.shape, (a, b))

    def test_get_reshsper(self):
        size_from, size_to = 10, 20

        # First time, it should create a new resherper
        resherper = gnn.quaziIdentity.get_reshsper(size_from, size_to)
        self.assertIsNotNone(resherper)
        self.assertEqual(resherper.shape, (size_from, size_to))

        # Second time, it should fetch the same one
        resherper_again = gnn.quaziIdentity.get_reshsper(size_from, size_to)
        self.assertTrue(np.array_equal(resherper, resherper_again))

        # Test identity scenario (no reshape needed)
        self.assertIsNone(gnn.quaziIdentity.get_reshsper(10, 10))

    def test_reshape(self):
        input_data = np.random.rand(10, 5)  # 10 features, 5 samples
        QIdentity = gnn.quaziIdentity.eye_stretch(10, 20)
        reshaped_data = gnn.quaziIdentity.Reshape(input_data, 20, QIdentity)

        self.assertEqual(reshaped_data.shape, (20, 5))
        self.assertTrue(np.allclose(np.dot(input_data[:, 0], QIdentity), reshaped_data[:, 0]))

    def test_reshape_forward_prop(self):
        input_data = np.random.rand(5, 3, 3, 2)  # Batch of 5, 3x3x2 feature maps
        flatten_size = 3 * 3 * 2
        QIdentity = gnn.quaziIdentity.eye_stretch(flatten_size, 10)
        reshaped_data = gnn.quaziIdentity.Reshape_forward_prop(input_data, 10, QIdentity)

        self.assertEqual(reshaped_data.shape, (10, 5))

    def test_reshape_back_prop(self):
        output_size, input_shape = 10, (3, 3, 2)
        E = np.random.rand(10, 5)
        QIdentity = gnn.quaziIdentity.eye_stretch(output_size, np.prod(input_shape))
        back_prop_data = gnn.quaziIdentity.Reshape_back_prop(E, input_shape, QIdentity)

        self.assertEqual(back_prop_data.shape, (5, 3, 3, 2))


if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import unittest
from testSuite import mode

class TestModel(unittest.TestCase):
    def setUp(self):
        global mode
        if mode == 'cpu':
            gnn.switch_to_cpu()
        elif mode == 'gpu':
            gnn.switch_to_gpu()
        self.model = gnn.Model(input_size=10, hidden_size=5, output_size=3)

    def test_init(self):
        self.assertEqual(self.model.batch_size, 128)
        self.assertEqual(self.model.input_size, 10)
        self.assertEqual(self.model.output_size, 3)
        self.assertEqual(self.model.hidden_size, 5)
        self.assertEqual(len(self.model.hidden_layers), 0)
        self.assertEqual(self.model.avaible_id, 2)
        self.assertEqual(self.model.convolution, False)
        self.assertIsNone(self.model.input_shape)
        self.assertIsNone(self.model.kernel_size)
        self.assertIsNone(self.model.depth)

    def test_forward_prop(self):
        input_data = gnn.np.random.rand(10, 10)  # Example input data
        result = self.model.forward_prop(input_data)
        self.assertEqual(result.shape, (3, 10))  # Assuming output_size is 3

    def test_add_and_remove_layer(self):
        self.assertEqual(len(self.model.hidden_layers), 0)
        layer_id = self.model.add_res_layer(layer_from_id='init_0', layer_to_id=1)
        self.assertEqual(len(self.model.hidden_layers), 1)
        self.model.remove_layer(layer_id)
        self.assertEqual(len(self.model.hidden_layers), 0)

if __name__ == '__main__':
    unittest.main()



import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
from numba import jit
import numpy as np
import cv2 as cv
from .helpers import *

RESHEPERS = {}

def eye_stretch(a,b):
    A = np.eye(max(a,b))
    return np.ascontiguousarray(np.array(cv.resize(get_numpy_array(A), (a,b))).T)

def get_reshsper(size_from, size_to):
    if size_from == size_to:
        return None
    elif not (size_from, size_to) in RESHEPERS.keys():
        RESHEPERS[(size_from, size_to)] = eye_stretch(size_from, size_to)
    return RESHEPERS[(size_from, size_to)]

def Reshape(x, output_size, QIdentity):
    x_reshaped = np.zeros((output_size, x.shape[1]))
    for i in range(0, x.shape[1]):
        if QIdentity is None:
            x_reshaped[:, i] = x[:, i]
        else:
            x_reshaped[:, i] = np.dot(x[:, i], QIdentity)
    return x_reshaped

def Reshape_forward_prop(x, output_size, QIdentity):
    x_reshaped = np.zeros((output_size, x.shape[0]))
    for i in range(0, x.shape[0]):
        flatten_size = x[i].shape[0] * x[i].shape[1] * x[i].shape[2]
        flatten = np.reshape(x[i], flatten_size)
        if QIdentity is None:
            x_reshaped[:, i] = flatten
        else:
            x_reshaped[:, i] = flatten.dot(QIdentity)
    return x_reshaped

def Reshape_back_prop(E, input_shape, QIdentity):
    E_reshaped = np.zeros((E.shape[1], input_shape[0], input_shape[1], input_shape[2]))
    for i in range(0, E.shape[1]):
        if QIdentity is None:
            x_reshaped = E[:, i]
        else:
            x_reshaped = E[:, i].dot(QIdentity)
        E_reshaped[i, :, :, :] = np.reshape(x_reshaped, input_shape)
    return E_reshaped



import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
from numba import jit
import cv2 as cv
import json
import random
import numpy

class NumpyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.int32):
            return int(obj)
        return json.JSONEncoder.default(self, obj)

def get_reverse_normal_distribution(clip_range, shape):
    tmp = np.random.normal(loc=0.0, scale=clip_range, size=shape)
    tmp_right = tmp + 2.2 * clip_range
    tmp_left = tmp - 2.2 * clip_range
    # Ensure we have enough samples for the desired shape
    num_samples = np.prod(shape)  # Total number of samples needed
    half_samples = num_samples // 2  # Half for left, half for right
    
    # Randomly choose samples for left and right sides
    left_selected = np.random.choice(tmp_left.flatten(), half_samples, replace=True)
    right_selected = np.random.choice(tmp_right.flatten(), half_samples, replace=True)
    
    # Merge left and right samples
    merged = np.concatenate((left_selected, right_selected))
    
    # Shuffle the merged array to avoid order bias
    np.random.shuffle(merged)
    
    # Reshape to the desired shape
    return merged.reshape(shape)

def switch_to_gpu():
    #print(" helper: switch_to_gpu")
    global np, IS_CUPY
    import cupy as np
    IS_CUPY = True

def switch_to_cpu():
    #print(" helper: switch_to_cpu")
    global np, IS_CUPY
    import numpy as np
    IS_CUPY = False
    
def clip(X, min, max):
    return np.array(fastclip(get_numpy_array(X), min, max))
    return np.array(np.clip(get_numpy_array(X), min, max))

@jit(nopython=True)
def fastclip(X : numpy, min : int, max : int):
    return np.clip(X, min, max)

def argmax(X, axis):
    return np.array(numpy.argmax(get_numpy_array(X), axis))

def randn(shape):
        return np.array(numpy.random.randn(shape))

def get_list_as_numpy_array(X):
    for i in range(0, len(X)):
        X[i] = get_numpy_array(X[i])
    return X
    
def get_numpy_array(X):
    if IS_CUPY == True:
        import cupy
        if isinstance(X, cupy.ndarray):
            return X.get()
        else:
            return numpy.array(X)
    else:
        return numpy.array(X)
    
def convert_to_desired_type(X):
    if not isinstance(X, np.ndarray):
        return np.array(X)
    return X
    
def one_hot(Y, Y_max = 0):
    Y_max = int(max(Y_max, Y.max() + 1))
    one_hot_Y = np.zeros((Y.size, Y_max))
    one_hot_Y[np.arange(Y.size), Y] = 1
    one_hot_Y = one_hot_Y.T
    return one_hot_Y

#@jit(nopython=True)
def add_n(array):
    sum = array[0]
    for i in range(1,len(array)): 
        sum += array[i]
    return sum

#@jit(nopython=True)
def mean_n(array):
    sum = add_n(array)
    div = float(len(array))
    return sum / div

#@jit(nopython=True)
def mean_n_conv(array, shape):
    sum = array[0]
    for i in range(1,len(array)): 
        sum += array[i]
    div = float(len(array))
    return sum / div

def delete_repetitions(array):
    result = []
    for obj in array:
        if not obj in result:
            result.append(obj)
    return result



def strech(x, shape):
    result = np.zeros((shape[0], shape[1], x.shape[2]))
    for i in range(0, x.shape[2]):
        result[:,:,i] = np.array(cv.resize(get_numpy_array(x[:,:,i]), shape))
    return result

def draw_hist(hist, label, path):
    xc = range(0, len(hist['train']))
    plt.figure()
    plt.plot(xc, hist['train'], label = "train acc")
    plt.plot(xc, hist['test'], label = "test acc")
    plt.legend()
    plt.savefig(path + "/" +label +".png")
    plt.close()

def get_max_loss(y):
    examples = 1
    for s in y.shape: examples *= s
    return examples * np.max(y)

def train_test_split_many_inputs(x, y, test_size):
    x = np.swapaxes(x, 0, 1)
    x_train = x[int(x.shape[0]* test_size) : x.shape[0]]
    x_test = x[0: int(x.shape[0] * test_size)]
    y_train = y[int(y.shape[0]* test_size) : y.shape[0]]
    y_test = y[0: int(y.shape[0] * test_size)] 

    x_train = np.swapaxes(x_train, 0, 1)
    x_test = np.swapaxes(x_test, 0, 1)
    return x_train, x_test, y_train, y_test

def protected_sampling(x, y, n):
    unique_classes = np.unique(y)
    num_classes = len(unique_classes)
    samples_per_class = max(1, n // num_classes)
    selected_indices = []
    for cls in unique_classes:
        class_indices = np.where(y == cls)[0]
        selected_indices.extend(np.random.choice(class_indices, size=min(samples_per_class, len(class_indices))))
    return select_data_at_indices(x, y, selected_indices)

def select_data_at_indices(x, y, selected_indices):
    matching_dim_x = None
    matching_dim_y = None # should be allways 0
    shape_x = list(x.shape)
    shape_y = list(y.shape)
    for i in range(0, len(shape_x)):
        for j in range(0, len(shape_y)):
            if shape_x[i] == shape_y[j]:
                matching_dim_x = i
                matching_dim_y = j

    if matching_dim_x is None:
        raise ValueError("Nie znaleziono odpowiedniego wymiaru zgodnego dla x i y.")
    x_selected = np.take(x, selected_indices, axis=matching_dim_x)
    y_selected = np.take(y, selected_indices, axis=matching_dim_y)

    return x_selected, y_selected

def limit_classes(x_train, y_train, x_test, y_test, num_classes=5):
    selected_indices_train = np.where(y_train < num_classes)
    selected_indices_test = np.where(y_test < num_classes)
    x_train = x_train[selected_indices_train]
    y_train = y_train[selected_indices_train]
    x_test = x_test[selected_indices_test]
    y_test = y_test[selected_indices_test]
    return x_train, y_train, x_test, y_test

def set_seed(new_seed):
    np.random.seed(new_seed)
    random.seed(new_seed)



import time
from ...structure import *
import networkx as nx

def to_networkx_graph(M):
    G = nx.DiGraph()
    for layer in M.input_layers + M.hidden_layers:
        G.add_node(layer.id)
        for input_layer_id in layer.input_layers_ids:
            G.add_edge(input_layer_id, layer.id)
        for output_layer_id in layer.output_layers_ids:
            G.add_edge(layer.id, output_layer_id)
    return G


def scoreDiameter(M, epochs, X_train, Y_train):
    # The smaller the normalized value, the smaller the graph's diameter, 
    # indicating nodes are more concentrated around a central point. Conversely, 
    # a larger value suggests a larger graph diameter, 
    # indicating nodes are more spread out.
    G = to_networkx_graph(M)
    max_distance = len(G.nodes()) - 1
    try:
        diameter = nx.diameter(G)
        normalized_diameter = diameter / max_distance
    except nx.NetworkXError:
        normalized_diameter = 1.0  # Graf niespójny, najgorszy przypadek
    return 1.0 - normalized_diameter

def scoreRadius(M, epochs, X_train, Y_train):
    # The smaller the normalized value, the smaller the graph's radius,
    # suggesting the graph is more centralized around a single node. 
    # A larger value indicates a larger graph radius, 
    # meaning the graph is more extensive.
    G = to_networkx_graph(M)
    max_distance = len(G.nodes()) - 1
    try:
        radius = nx.radius(G)
        normalized_radius = radius / max_distance
    except nx.NetworkXError:
        normalized_radius = 1.0  # Graf niespójny, najgorszy przypadek
    return 1.0 - normalized_radius

def scoreEccentricity(M, epochs, X_train, Y_train):
    # The smaller the normalized value, the smaller the eccentricity of nodes, 
    # indicating nodes are closer together and more integrated. 
    # A larger value suggests greater eccentricity of nodes, 
    # meaning nodes are more distant from each other.
    G = to_networkx_graph(M)
    max_distance = len(G.nodes()) - 1
    # Eccentricity
    try:
        eccentricity = nx.eccentricity(G)
        normalized_eccentricity = {n: e / max_distance for n, e in eccentricity.items()}
    except nx.NetworkXError:
        normalized_eccentricity = 0.0  # Graf niespójny, najgorszy przypadek
    return normalized_eccentricity


def scoreChromaticNumber(M, epochs, X_train, Y_train):
    # The smaller the normalized value, the greater the chromatic number, 
    # indicating the graph requires more distinct colors to color the graph without conflicts. 
    # Conversely, a larger value suggests a smaller chromatic number, 
    # meaning the graph requires fewer distinct colors.
    G = to_networkx_graph(M)
    chromatic_number = nx.coloring.greedy_color(G.to_undirected(), strategy='largest_first')
    max_chromatic_number = max(chromatic_number.values()) + 1
    normalized_chromatic_number = 1 / max_chromatic_number  
    return normalized_chromatic_number


def scoreMatchingNumber(M, epochs, X_train, Y_train):
    # The smaller the normalized value, the greater the matching number, 
    # indicating more edges of the graph can be paired without conflict. 
    # A larger value suggests a smaller matching number, 
    # meaning fewer edges can be paired.
    G = to_networkx_graph(M)
    matching_number = len(nx.max_weight_matching(G.to_undirected()))
    normalized_matching_number = 1 / matching_number if matching_number > 0 else 1.0  
    return normalized_matching_number

def scoreIndependenceNumber(M, epochs, X_train, Y_train):
    # The smaller the normalized value, the greater the independence number, 
    # indicating more vertices of the graph can be independent (not connected by edges).
    # A larger value suggests a smaller independence number, 
    # meaning fewer vertices can be independent.
    G = to_networkx_graph(M)
    independence_number = len(nx.maximal_independent_set(G.to_undirected()))
    normalized_independence_number = 1 / independence_number if independence_number > 0 else 1.0  
    return normalized_independence_number


import math
import sys
from enum import Enum
import json
import threading
import os
import time
from .painter import *
from .config import *
from .optimizers import *
from .quaziIdentity import *

def switch_to_gpu():
    global np, IS_CUPY, correlate, convolve
    import cupy as np
    IS_CUPY = True
    from cupyx.scipy.ndimage import correlate
    from cupyx.scipy.ndimage import convolve

def switch_to_cpu():
    global np, IS_CUPY, correlate2d, convolve2d
    import numpy as np
    IS_CUPY = False
    from scipy.signal import correlate2d
    from scipy.signal import convolve2d

class Loss:
    
    def getByName(name):
        if name == Loss.MSE.__name__:
            return Loss.MSE
        elif name == Loss.multiclass_cross_entropy.__name__:
            return Loss.multiclass_cross_entropy

    class MSE:
        __name__ = 'MSE'
        def exe(Y_true, Y_pred):
            return np.sum((Y_pred - Y_true)**2)/Y_pred.shape[0]
        def der(Y_true, Y_pred):
            return Y_pred - Y_true
    class multiclass_cross_entropy:
        __name__ = 'multiclass_cross_entropy'
        def exe(Y_true, Y_pred):
            error = 0.0
            for i in range(0, Y_true.shape[1]):
                error -= np.dot(Y_true[:,i].T, np.log(Y_pred[:,i]))
            #print("Y_true.shape: ", Y_true.shape)
            return error / Y_true.shape[1]
        def der(Y_true, Y_pred):
            grad = np.zeros(Y_true.shape)
            for i in range(0, Y_true.shape[1]):
                #print("Y_true[:, i]: ", Y_true[:, i].shape)
                #print("Y_pred[:, i]: ", Y_pred[:, i].shape)
                partial_grad = -Y_true[:, i] / Y_pred[:, i]
                A = np.tile(np.reshape(Y_pred[:, i], (Y_pred.shape[0], 1)), (1, Y_pred.shape[0]))
                grad[:, i] = (A * (np.identity(Y_pred.shape[0]) - A.T)) @ partial_grad
            return grad
               
class Activations:

    def getByName(name):
        if name == Activations.ReLu.__name__:
            return Activations.ReLu
        elif name == Activations.leaky_ReLu.__name__:
            return Activations.leaky_ReLu
        elif name == Activations.SoftMax.__name__:
            return Activations.SoftMax
        elif name == Activations.Sigmoid.__name__:
            return Activations.Sigmoid
    class ReLu:
        __name__ = 'ReLu'

        @staticmethod
        @jit(nopython=True)
        def exe(X):
            return numpy.maximum(X,0)
        
        @staticmethod
        @jit(nopython=True)
        def der(X):
            return X > 0
        
    class leaky_ReLu:
        __name__ = 'leaky_ReLu'

        @staticmethod
        @jit(nopython=True)
        def exe(X):
            return  np.where(X > 0, X, X * 0.001)
        
        @staticmethod
        @jit(nopython=True)
        def der(X):
            return  np.where(X > 0, 1, 0.001)
        
    class SoftMax:
        __name__ = 'SoftMax'

        @staticmethod
        def exe(X):
            result = np.zeros(X.shape)
            for i in range(0, X.shape[1]):
                exp = np.exp(X[:, i] - np.nanmax(X[:, i]))
                result[:, i] = np.nan_to_num(exp / np.sum(exp))
            return clip(result, 0.0001, 0.999)
        
        @staticmethod
        @jit(nopython=True)
        def der(X):
            return 1.0
        
    class Sigmoid:
        __name__ = 'Sigmoid'

        @staticmethod
        @jit(nopython=True)
        def exe(X):
            return  1/(1 + np.exp(-X))
        
        @staticmethod
        @jit(nopython=True)
        def der(X):
            sigm = 1/(1 + np.exp(-X))
            return sigm * (1.0 - sigm)
        
class LearningRateScheduler:
    CONSTANT = 0
    PROGRESIVE = 1
    PROGRESIVE_PARABOIDAL = 2

    def __init__(self, mode, alpha, steepness = 0.2):
        self.mode = mode
        self.alpha = alpha
        self.steepness = steepness

    def alpha_scheduler(self, i, iterations):
        thresh = float(self.steepness * iterations)
        i = float(i)
        iterations = float(iterations)
        result = self.alpha
        if self.mode == LearningRateScheduler.CONSTANT:
            result =  self.alpha
        elif self.mode == LearningRateScheduler.PROGRESIVE:
            if i < thresh:
                return self.alpha * ((i+1) / (thresh + 2))
            result = self.alpha * (1 - (i - thresh) / (iterations - thresh + 2))
        else:
            thresh = self.steepness * iterations
            if i < thresh:
                return self.alpha * ( -1 * ((1) / (pow(thresh, 2))) * pow((i) - thresh, 2) + 1)
            result = self.alpha * (-1 * ((1) / (pow(iterations - thresh, 2))) * pow((i) - thresh, 2) + 1)
        return max(0, result)

class SimulationScheduler:
    CONSTANT = 0
    PROGRESS_CHECK = 1
    OVERFIT_CHECK = 2

    def __init__(self, mode, simulation_time, simulation_epochs, min_grow_rate = 0.65):
        self.mode = mode
        self.simulation_time = simulation_time
        self.simulation_epochs = simulation_epochs
        self.min_grow_rate = min_grow_rate
        self.flatten_check_length = 20
        self.progres_delta = 0.01

    def can_simulate(self, i, hist_detail):
        new_acc = 0#hist_detail.Y['iteration_acc_train'][-1]
        prev_acc = 0#hist_detail.Y['iteration_acc_train'][-2]
        if len(hist_detail.Y['iteration_acc_train']) > 0: new_acc = hist_detail.Y['iteration_acc_train'][-1]
        if len(hist_detail.Y['iteration_acc_train']) > 1: prev_acc = hist_detail.Y['iteration_acc_train'][-2]
        new_acc = get_numpy_array(new_acc)
        prev_acc = get_numpy_array(prev_acc)
        if self.mode == SimulationScheduler.CONSTANT: 
            print("[iteration: "+str(i)+"] Constant frequency od simulaiton acc: " + str(new_acc)+ " starting simulation." )
            return True
        elif self.mode == SimulationScheduler.PROGRESS_CHECK:
            if new_acc - prev_acc < self.progres_delta:
                print("[iteration: "+str(i)+"] No correction detected acc: " + str(new_acc)+ "(prev: " + str(prev_acc) + ") starting simulation."  )
                return True
            else:
                print("[iteration: "+str(i)+"] correction detected acc: " + str(new_acc)+ "(prev: " + str(prev_acc) + ") training continues.")
                return False
        elif self.mode == SimulationScheduler.OVERFIT_CHECK:
            if hist_detail.learning_capable():
                print("[iteration: "+str(i)+"] Overfit detected: " + str(new_acc)+ " starting simulation."  )
                return True
            else:
                print("[iteration: "+str(i)+"] No overfit detected: " + str(new_acc)+ " training continues.")
                return False
        return False
    
    def get_mode_label(self):
        if self.mode == SimulationScheduler.CONSTANT: return "simconstant"
        elif self.mode == SimulationScheduler.PROGRESS_CHECK: return "simprogres"
        elif self.mode == SimulationScheduler.OVERFIT_CHECK: return "simoverfit"


class History:
    def __init__(self, keys):
        self.Y = {}
        self.last_img_id = 0
        self.description = "\ndescription_of_training_process: \n"
        self.best_train_acc = 0.0
        self.best_test_acc = 0.0
        for key in keys:
            self.Y[key] = []
    def get_length(self):
        return len(self.Y[list(self.Y.keys())[0]])
    def append(self, key, value):
        if not key in self.Y.keys():
            self.Y[key] = []
        self.Y[key].append(value)
    def merge(self, new_hist):
        for key in self.Y.keys():
            if key in self.Y.keys() and key in new_hist.Y.keys():
                self.Y[key] += new_hist.Y[key]
    def learning_capable(self, patience=10, verbose=0.5):
        if len(self.Y['loss']) < patience: return True
        diff = 0
        last_values = self.Y['loss'][::-1]
        for i in range(1, len(last_values)):
            diff += abs(last_values[i] - last_values[i - 1])
        return diff > verbose
    def get_last(self, key):
        return self.Y[key][-1]
    def draw_hist(self, label, path):
        for key in self.Y.keys():
            #print("len(self.Y[key]): ", len(self.Y[key]), type(self.Y[key][0]))
            xc = range(0, len(self.Y[key]))
            plt.figure()
            plt.plot(xc, get_list_as_numpy_array(self.Y[key]), label = key)
            plt.legend()
            plt.savefig(path + "/" + label + "_" + key + ".png")
            plt.close()
    def save(self, path):
        dict = {}
        dict['keys'] = {}
        for key in self.Y.keys():
            dict['keys'][key] = np.array(get_list_as_numpy_array(self.Y[key]))
        dict['last_img_id'] = str(self.last_img_id)
        dict['description'] = str(self.description)
        dict['best_train_acc'] = str(self.best_train_acc)
        dict['best_test_acc'] = str(self.best_test_acc)
        with open(path, 'w+') as f:
            f.write(json.dumps(dict, cls=NumpyArrayEncoder))
        with open(path+"_description.txt", 'w') as f:
            f.write(self.description)
    def load(self, path):
        with open(path, "r") as f:
            data = json.load(f)
        self.Y = {}
        self.last_img_id = int(data['last_img_id'])
        self.description = str(data['description'])
        if 'best_train_acc' in data.keys():
            self.best_train_acc = float(data['best_train_acc'])
            self.best_test_acc = float(data['best_test_acc'])
        for key in data['keys'].keys():
            self.Y[key] = list(np.asarray(data['keys'][key]))

class Layer_Type(Enum):
    ZERO = 1
    RANDOM = 2
    EYE = 3

class Layer:
    def __init__(self, _id, _model, input_size, neurons, act_fun, layer_type = Layer_Type.RANDOM, _optimizer = SGDOptimizer()):
        self.id = _id
        self.model = _model
        self.input_size = input_size
        self.neurons = neurons
        self.act_fun = act_fun
        self.is_ending = False
        self.is_starting = False
        self.input_layers_ids = []
        self.output_layers_ids = []
        self.f_input = [] # forwarind input
        self.b_input = [] # backwarding input
        self.reshspers = {}
        self.optimizer = _optimizer
        self.optimizer_W = OptimizerFactory.copy(self.optimizer)
        self.optimizer_B = OptimizerFactory.copy(self.optimizer)
        self.connections = {}
        if layer_type == Layer_Type.EYE:
            self.W = np.asarray(eye_stretch(neurons, input_size))
            self.B = np.asarray(np.zeros((neurons, 1)))
        elif layer_type == Layer_Type.ZERO:
            self.W = np.asarray(np.zeros((neurons, input_size)))
            self.B = np.asarray(np.zeros((neurons, 1)))
        else:
            if WEIGHT_DISTRIBUTION_MODE == DistributionMode.UNIFORM:
                self.W = np.random.uniform(low=-WEIGHTS_CLIP_RANGE/3, high=WEIGHTS_CLIP_RANGE/3, size=(neurons, input_size))
                self.B = np.random.uniform(low=-WEIGHTS_CLIP_RANGE/3, high=WEIGHTS_CLIP_RANGE/3, size=(neurons, 1))
            elif WEIGHT_DISTRIBUTION_MODE == DistributionMode.NORMAL:
                self.W = np.random.normal(loc=0.0, scale=WEIGHTS_CLIP_RANGE/3, size=(neurons, input_size))
                self.B = np.random.normal(loc=0.0, scale=WEIGHTS_CLIP_RANGE/3, size=(neurons, 1))
            elif WEIGHT_DISTRIBUTION_MODE == DistributionMode.GAMMA:
                # Using shape=2 and scale=WEIGHTS_CLIP_RANGE for Gamma distribution.
                self.W = np.random.gamma(shape=2.0, scale=WEIGHTS_CLIP_RANGE/3, size=(neurons, input_size))
                self.B = np.random.gamma(shape=2.0, scale=WEIGHTS_CLIP_RANGE/3, size=(neurons, 1))
            elif WEIGHT_DISTRIBUTION_MODE == DistributionMode.REVERSED_GAUSSIAN:
                # Shifting the mean to negative, and controlling spread with WEIGHTS_CLIP_RANGE
                self.W = get_reverse_normal_distribution(WEIGHTS_CLIP_RANGE/3, (neurons, input_size))
                self.B = get_reverse_normal_distribution(WEIGHTS_CLIP_RANGE/3, (neurons, 1))
            else:
                raise ValueError(f"Unsupported distribution mode: {WEIGHT_DISTRIBUTION_MODE}")
            
            
    def set_as_ending(self):
        self.is_ending = True
        #self.done_event = threading.Event()
    
    def set_as_starting(self):
        self.is_starting = True
        #self.done_event = threading.Event()
        
    def get_output_size(self):
        return self.neurons
    
    def connect_input(self, layer_id):
        if layer_id == self.id: 
            print("error I")
            return
        target_layer = self.model.get_layer(layer_id)
        if not layer_id in self.input_layers_ids:
            self.input_layers_ids.append(layer_id)
        if not self.id in target_layer.output_layers_ids:
            target_layer.connect_output(self.id)

    def connect_output(self, layer_id):
        if layer_id == self.id: 
            print("error O")
            print("self.id: "+ str(self.id)+ " layer_id: "+ str(layer_id))
            return
        target_layer = self.model.get_layer(layer_id)
        if not layer_id in self.output_layers_ids:
            self.output_layers_ids.append(layer_id)
        if not self in target_layer.input_layers_ids:
            target_layer.connect_input(self.id)
    
    def disconnect(self, to_remove_layer_id):
        if to_remove_layer_id in self.input_layers_ids:
            self.input_layers_ids.remove(to_remove_layer_id)
        if to_remove_layer_id in self.output_layers_ids:
            self.output_layers_ids.remove(to_remove_layer_id)
        if self.id == to_remove_layer_id:
            self.input_layers_ids = []
            self.output_layers_ids = []

    @staticmethod
    @jit(nopython=True)
    def update_weights_shape(W, input_size):
        current_weight_size = W.shape[1]
        if current_weight_size < input_size:
            # Dodawanie brakujących kolumn z zerami
            extra_columns = input_size - current_weight_size
            zero_padding = np.zeros((W.shape[0], extra_columns))
            # Zamiast np.hstack używamy manualnej konkatenacji
            W = np.concatenate((W, zero_padding), axis=1)
        elif current_weight_size > input_size:
            # Usuwanie niepotrzebnych kolumn
            W = W[:, :input_size]
        return W
    
    def should_thread_forward(self):
        if threading.active_count() >= MAX_THREADS:
            return False
        if len(self.f_input) + 1 < len(self.input_layers_ids): 
            return False
        return True
    
    def append_to_f_input(self, X, sender_id):
        if sender_id == -1:
            self.f_input = [X]
        else:
            if sender_id not in self.input_layers_ids:
                #return None
                raise ValueError(f"Sender ID {sender_id} is not in the input layers IDs {self.input_layers_ids}")
            while len(self.f_input) < len(self.input_layers_ids):
                self.f_input.append(None)  # Uzupełniamy miejsce, jeśli f_input jest krótsze niż input_layers_ids
            self.f_input[self.input_layers_ids.index(sender_id)] = X
        
    def forward_prop(self, X, sender_id, deepth = 0):
        #self.f_input.append(X)
        self.append_to_f_input(X, sender_id)
        if any(x is None for x in self.f_input):
                return None
        #if len(self.f_input) < len(self.input_layers_ids): 
        #    return None
        input_list = []
        for layer_input in self.f_input:
            input_list.append(layer_input)  # Zbiera dane wejściowe
        self.I = np.vstack(input_list)
        self.W = Layer.update_weights_shape(self.W, self.I.shape[0])
        self.Z = np.dot(self.W, self.I)
        self.A = self.act_fun.exe(self.Z)

        for layer_id in self.output_layers_ids:
            layer = self.model.get_layer(layer_id)
            if type(layer) == Layer:
                new_input = Reshape(self.A.copy(), layer.input_size, get_reshsper(self.A.shape[0], layer.input_size))

            #TODO: Check if creating a thread makes sense
            #TODO: new_input CAN BE SEND BEFORE ASSIGMENT WHY ?
            if layer.should_thread_forward():
                thread = threading.Thread(
                    target=lambda: self.model.get_layer(layer_id).forward_prop(new_input, self.id, deepth + 1),
                )
                thread.start()
                self.model.forward_threads.append(thread)
            else:
                #print(f"No available threads, continuing in the current thread: {threading.current_thread().name}")
                self.model.get_layer(layer_id).forward_prop(new_input, self.id, deepth + 1)
        self.f_input = []


    def should_thread_backward(self):
        if threading.active_count() >= MAX_THREADS:
            return False
        if len(self.b_input) + 1 < len(self.output_layers_ids): 
            return False
        return True
    
    def back_prop(self,E,m,alpha):
        if E.shape[0] <=0:
            raise ValueError("Error with 0 shape can't be backpropagated E.shape:", E.shape)
        m = 1.0
        E = Reshape(E, self.neurons, get_reshsper(E.shape[0], self.neurons))
        self.b_input.append(E)
        if len(self.b_input) < len(self.output_layers_ids): return None
        self.E =  clip(mean_n(self.b_input), -error_clip_range, error_clip_range)
        dZ = self.E * self.act_fun.der(self.Z)
        self.dW = Layer.calcuale_dW(m, dZ, self.I)
        self.dB = Layer.calcuale_dB(m, dZ, self.B)
        self.E = self.W.T @ dZ
        before_iteration = 0
        for layer_id in self.input_layers_ids:
            neurons = self.input_size
            E_slice = self.W[:, before_iteration:before_iteration + neurons].T @ dZ
            before_iteration += neurons
            layer = self.model.get_layer(layer_id)
            if layer.should_thread_backward():
                thread = threading.Thread(
                    target=lambda: layer.back_prop(E_slice.copy(), m, alpha),
                )
                thread.start()
                self.model.bacward_threads.append(thread)
            else:
                #print(f"No available threads, continuing in the current thread: {threading.current_thread().name} count: {threading.active_count()}")
                layer.back_prop(E_slice, m, alpha)
        self.update_params(alpha)
        self.b_input = []
        # if self.is_starting:
        #     self.done_event.set()

    def update_params(self, alpha):
        self.W = self.optimizer_W.update(self.W, self.dW, alpha)
        self.B = self.optimizer_B.update(self.B, self.dB, alpha)

    @staticmethod
    @jit(nopython=True)
    def calcuale_Z(W, I, B):
        return np.dot(W, I) + B

    @staticmethod
    @jit(nopython=True)
    def calcuale_dW(m, dZ, I):
        return 1 / m * dZ @ I.T

    def calcuale_dB(m, dZ, B):
        return 1 / m * np.reshape(np.sum(dZ, 1), B.shape)
    
    @staticmethod
    @jit(nopython=True)
    def calcuale_updateW(W, alpha, dw):
        return W - alpha * dw
    

    def get_all_childrens_connections(self, deepth = 0):
        conn = []
        for layer_id in self.output_layers_ids:
            conn.append([self.id, layer_id])
            conn += self.model.get_layer(layer_id).get_all_childrens_connections(deepth + 1)
        return conn

    def is_cyclic(self, visited, additional_pair, depth = 0):
        depth += 1
        to_check = self.output_layers_ids.copy()
        if additional_pair[0] == self.id:
            to_check.append(additional_pair[1])
        for layer_id in to_check:
            if layer_id in visited: return True
            visited.append(layer_id)
            layer = self.model.get_layer(layer_id)
            cycle_found = layer.is_cyclic(visited, additional_pair, depth)
            if cycle_found: return True
            visited.remove(layer_id)
        return False
        
    def deepcopy(self):
        copy = Layer(self.id, None, self.input_size, self.neurons, self.act_fun, Layer_Type.RANDOM, self.optimizer.getDense())
        if self.is_ending:
            copy.set_as_ending()
        if self.is_starting:
            copy.set_as_starting()
        copy.input_layers_ids = self.input_layers_ids.copy()
        copy.output_layers_ids = self.output_layers_ids.copy()
        copy.f_input = self.f_input.copy()
        copy.b_input = self.b_input.copy()
        copy.W = self.W.copy()
        copy.B = self.B.copy()
        return copy

    def get_weights_summary(self):
        return "W: "+str(np.mean(self.W))+ " B: "+ str(np.mean(self.B))
    
    def get_paint_label(self):
            # self.W.shape[1] -> connections/input size
            # self.W.shape[0] -> neurons/outpu size
            return str(self.id) + "[" +str(self.W.shape[1])+","+str(self.W.shape[0])+"]"

    def __str__(self):
        return "[<layer: "+ str(self.id)+ " id: " + str(id(self)) + " model id: "+ str(id(self.model))+" in conn: "+ str(len(self.input_layers_ids)) +" out conn: "+ str(len(self.output_layers_ids))+ ">]" 

class Model:
    def __init__(self, input_size, hidden_size, output_size, loss_function = Loss.multiclass_cross_entropy, activation_fun = Activations.Sigmoid, input_paths = 1, _optimizer = SGDOptimizer()):
        self.batch_size = 128
        self.loss_function = loss_function
        self.input_size = input_size
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.hidden_layers = []
        self.avaible_id = 2
        self.activation_fun = activation_fun
        self.input_layers = []#Layer(0, self, input_size, hidden_size, self.activation_fun)
        self.optimizer = _optimizer
        self.output_layer = Layer(1, self, hidden_size, output_size, Activations.SoftMax, Layer_Type.RANDOM, self.optimizer.getDense())
        self.output_layer.set_as_ending()
        for i in range(0, input_paths):
            layer_id = "init_"+str(i)
            layer = Layer(layer_id, self, input_size, hidden_size, self.activation_fun, Layer_Type.RANDOM, self.optimizer.getDense())
            layer.set_as_starting()
            self.input_layers.append(layer)
            self.add_connection(layer_id, self.output_layer.id)
        if input_paths > 1: self.add_sequential_output_Layer()
        # in testing:
        self.convolution = False
        self.input_shape = None
        self.kernel_size = None
        self.depth = None
        self.forward_threads = []
        self.bacward_threads = []
        

    def set_convolution_mode(self, input_shape, kernel_size, depth):
        #import convolution
        self.convolution = True
        self.input_shape = input_shape
        self.kernel_size = kernel_size
        self.depth = depth # amount of kernels
        for i in range(0, len(self.input_layers)):
            layer_id = "init_"+str(i)
            output_layers_ids = self.input_layers[i].output_layers_ids
            for output_layer_id in output_layers_ids:
                self.input_layers[i].disconnect(output_layer_id)
                self.get_layer(output_layer_id).disconnect(self.input_layers[i].id)
                self.input_layers[i] = Conv(layer_id, self, self.input_shape, self.kernel_size, self.depth, self.activation_fun, self.optimizer.getConv())
                self.add_connection(layer_id, output_layer_id)

    def add_res_layer(self, layer_from_id, layer_to_id, layer_type = Layer_Type.ZERO):
        layer_from = self.get_layer(layer_from_id)
        layer_to = self.get_layer(layer_to_id)
        input_size = layer_from.get_output_size()
        #input_size = min(input_size, self.hidden_size)
        new_layer = Layer(self.avaible_id, self, input_size, layer_to.input_size, self.activation_fun, layer_type, self.optimizer.getDense())
        self.hidden_layers.append(new_layer)
        self.add_connection(layer_from_id, new_layer.id)
        self.add_connection(new_layer.id, layer_to_id)
        self.avaible_id += 1
        return new_layer.id
    
    def add_norm_layer(self, layer_from_id, layer_to_id, layer_type = Layer_Type.RANDOM):
        layer_from = self.get_layer(layer_from_id)
        layer_to = self.get_layer(layer_to_id)
        if type(layer_from) == Conv:
            input_size = layer_from.output_flatten
        elif type(layer_from) == Layer:
            input_size = layer_from.neurons
        #input_size = min(input_size, self.hidden_size)
        new_layer = Layer(self.avaible_id, self, input_size, layer_to.input_size, self.activation_fun, layer_type, self.optimizer.getDense())
        self.hidden_layers.append(new_layer)
        self.add_connection(layer_from_id, new_layer.id)
        self.add_connection(new_layer.id, layer_to_id)
        layer_from.disconnect(layer_to_id)
        layer_to.disconnect(layer_from_id)
        self.avaible_id += 1
        return new_layer.id
    
    def add_sequential_output_Layer(self):
        new_layer = Layer(self.avaible_id, self, self.output_layer.input_size, self.output_layer.input_size, self.activation_fun, Layer_Type.EYE, self.optimizer.getDense())
        self.hidden_layers.append(new_layer)
        input_layers_ids = self.output_layer.input_layers_ids.copy()
        for input_layer_id in input_layers_ids:
            self.add_connection(input_layer_id, new_layer.id)
            layer_from = self.get_layer(input_layer_id)
            layer_from.disconnect(self.output_layer.id)
            self.output_layer.disconnect(layer_from.id)
        self.add_connection(new_layer.id, self.output_layer.id)
        self.avaible_id += 1
        return new_layer.id
    
    def add_conv_norm_layer(self, layer_from_id, layer_to_id, layer_type = Layer_Type.RANDOM):
        layer_from = self.get_layer(layer_from_id)
        layer_to = self.get_layer(layer_to_id)
        if type(layer_to) == Conv:
            i = layer_from.output_shape[0]
            d = clip(layer_to.depth, 1, 3)
            k = clip(layer_to.kernel_size, 1, i)
            new_layer = Conv(self.avaible_id, self, layer_from.output_shape, k, d, self.activation_fun, self.optimizer.getConv())
        elif type(layer_to) == Layer:
            o = layer_to.input_size
            i = layer_from.output_shape[0]
            d = clip(math.floor(o ** ( 1 / 3)), 1, 3)
            k = clip(i + 1 - math.ceil(o ** ( 1 / 3)), 1, i)
            new_layer = Conv(self.avaible_id, self, layer_from.output_shape, k, d, self.activation_fun, self.optimizer.getConv())
        self.hidden_layers.append(new_layer)
        self.add_connection(layer_from_id, new_layer.id)
        self.add_connection(new_layer.id, layer_to_id)
        layer_from.disconnect(layer_to_id)
        layer_to.disconnect(layer_from_id)
        self.avaible_id += 1
        return new_layer.id
    
    def add_conv_res_layer(self, layer_from_id, layer_to_id, layer_type = Layer_Type.ZERO):
        layer_from = self.get_layer(layer_from_id)
        layer_to = self.get_layer(layer_to_id)
        if type(layer_to) == Conv:
            d = layer_to.depth
            i = layer_from.output_shape[0]
            k = clip(layer_to.kernel_size, 1, i)
            new_layer = Conv(self.avaible_id, self, layer_from.output_shape, k, d, self.activation_fun, self.optimizer.getConv())        
        elif type(layer_to) == Layer:
            o = layer_to.input_size
            i = layer_from.output_shape[0]
            d = clip(math.floor(o ** ( 1 / 3)), 1, 3)
            k = clip(i + 1 - math.ceil(o ** ( 1 / 3)), 1, i)
            new_layer = Conv(self.avaible_id, self, layer_from.output_shape, k, d, self.activation_fun, self.optimizer.getConv())
        self.hidden_layers.append(new_layer)
        self.add_connection(layer_from_id, new_layer.id)
        self.add_connection(new_layer.id, layer_to_id)
        self.avaible_id += 1
        return new_layer.id
    
    def add_connection(self, layer_from_id, layer_to_id):
        L_from = self.get_layer(layer_from_id)
        L_from.connect_output(layer_to_id)

    def get_layer(self, id):
        for input_layer in self.input_layers:
            if input_layer.id == id: 
                return input_layer
        if self.output_layer.id == id: 
            return self.output_layer
        for layer in self.hidden_layers:
            if layer.id == id:
                return layer
            
    def get_predictions(A2):
        return argmax(A2, 0)

    def get_accuracy(predictions, Y):
        if predictions.shape != Y.shape:
            sys.exit("ERROR, shape of predictions is diffrent than one_hot_Y: " + str(predictions.shape) + " != " + str(Y.shape))
        return np.sum(predictions == Y) / Y.size

    def forward_prop(self, input):
        if not isinstance(input, np.ndarray):
            input = np.array(input)
        #print("len(self.input_layers): ", len(self.input_layers))
        self.output_layer.A = None
        self.output_layer.set_as_ending()
        if len(self.input_layers) == 1:
            self.input_layers[0].forward_prop(input, -1,  0)
        else:
            for i in range(0, len(self.input_layers)):
                self.input_layers[i].forward_prop(input[i], -1,  0)
        
        #self.output_layer.done_event.wait()
        for thread in self.forward_threads:
            thread.join()
        self.forward_threads.clear()
        if self.output_layer.A is None:
            raise ValueError("After forward prop A on output layer is None")
        return self.output_layer.A
        #return result
        
    #def back_prop(self,
    def back_prop(self,E,m,alpha):
        for i in range(0, len(self.input_layers)):
            self.input_layers[i].set_as_starting()
        self.output_layer.back_prop(E, m, alpha)   
        # for i in range(0, len(self.input_layers)):
        #     self.input_layers[i].done_event.wait()
        for thread in self.bacward_threads:
            thread.join()
        self.bacward_threads.clear()
    
    def gradient_descent(self, X, Y, iterations, lr_scheduler, quiet = False, one_hot_needed = True, path="."):
        if not isinstance(X, np.ndarray): X = np.array(X)
        if not isinstance(Y, np.ndarray): Y = np.array(Y)
        if one_hot_needed: one_hot_Y = one_hot(Y)
        else: one_hot_Y = Y
        A = []
        index_axis = 0
        if self.convolution:
            if len(self.input_layers) == 1:
                index_axis = 0
            else:
                index_axis = 1
        else:
            if len(self.input_layers) == 1:
                index_axis = 1
            else:
                index_axis = 2
        indexes = list(range(0, X.shape[index_axis]))
        history = History(['accuracy', 'loss'])
        for i in range(iterations + 1):
            current_alpha = lr_scheduler.alpha_scheduler(i, iterations) #alpha * Model.alpha_scheduler(i, iterations)
            loss = 0
            predictions = []  
            expectations = []
            for x_indx_start in range(0, X.shape[index_axis], self.batch_size): #lwn(x)
                batch_indexes = indexes[x_indx_start:(x_indx_start + self.batch_size)]
                A = self.forward_prop(np.take(X, batch_indexes, index_axis))
                E = self.loss_function.der(np.take(one_hot_Y, batch_indexes, 1) , A)
                self.back_prop(E, self.batch_size, current_alpha)
                loss += self.loss_function.exe(np.take(one_hot_Y, batch_indexes, 1) , A)
                predictions.append(A)
                expectations.append(np.take(Y, batch_indexes, 0))
            random.shuffle(indexes)
            #Storage.saveModel(self, path + "epoch_" + str(i) + "save.json")
            #draw(self, path + "epoch_" + str(i) +".html")
           
            acc = Model.get_accuracy(Model.get_predictions(np.concatenate(predictions, axis=1)), np.concatenate(expectations, axis=0))
            history.append('accuracy', acc)
            history.append('loss', loss)
            if i % 1 == 0 and quiet==False:
                print("Epoch: "+ str(i) + " Accuracy: " + str(round(float(acc),3))+ " loss: " + str(round(float(loss),3)) + " lr: " + str(round(float(current_alpha), 3)) +  " threads: " + str(threading.active_count()))
        #TODO: this forward seems useless
        #A = self.forward_prop(X)
        return history.get_last('accuracy'), history

    def evaluate(self, x, y):
        A = self.forward_prop(x)
        return Model.get_accuracy(Model.get_predictions(A),y)

    def remove_layer(self, layer_id, preserve_flow = True):
        if preserve_flow:
            layer_to_remove = self.get_layer(layer_id)
            for input_layer_id in layer_to_remove.input_layers_ids:
                input_layer = self.get_layer(input_layer_id)
                for output_layer_id in layer_to_remove.output_layers_ids:
                    input_layer.connect_output(output_layer_id)
        
        self.output_layer.disconnect(layer_id)
        for input_layer in self.input_layers:
            input_layer.disconnect(layer_id)
        for layer in self.hidden_layers:
            layer.disconnect(layer_id)
        layer_to_remove = self.get_layer(layer_id)
        self.hidden_layers.remove(layer_to_remove)

    def get_all_childrens_connections(self):
        pairs = []
        for input_layer in self.input_layers:
            pairs += input_layer.get_all_childrens_connections()
        return delete_repetitions(pairs)

    def deepcopy(self):
        copy = Model(self.input_size, self.output_size, self.hidden_size)
        copy.input_layers = []
        for input_layer in self.input_layers:
            copy_layer = input_layer.deepcopy()
            copy_layer.model = copy
            copy.input_layers.append(copy_layer)
        for layer in self.hidden_layers:
            copy_layer = layer.deepcopy()
            copy_layer.model = copy
            copy.hidden_layers.append(copy_layer)
        copy.avaible_id = self.avaible_id
        copy.output_layer = self.output_layer.deepcopy()
        copy.output_layer.model = copy
        copy.convolution = self.convolution
        copy.input_shape = self.input_shape
        copy.kernel_size = self.kernel_size
        copy.depth = self.depth
        return copy
    
    def is_cyclic(self, additional_pair):
        return self.input_layer.is_cyclic([], additional_pair)
 
    def is_connected(self, layer_to_del):
        return self.input_layer.is_connected(layer_to_del)

    def get_sequence_connection(self):
        pairs = []
        for layer in self.hidden_layers + self.input_layers:
            for input_id in layer.input_layers_ids:
                pairs.append([input_id, layer.id])
            for input_id in layer.output_layers_ids:
                pairs.append([layer.id, input_id])
        return pairs
    
    def get_state_representation(self):
        #experimental / temporary
        return str(self.get_sequence_connection())
    
    def show_connection_table(self):
        print("connection_table")
        for layer in self.input_layers + [self.output_layer]:
            print("layer.id: ", layer.id, " layer outputs: ", layer.output_layers_ids, " summary: ", layer.get_weights_summary())
        for layer in self.hidden_layers:
            print("layer.id: ", layer.id, " layer outputs: ", layer.output_layers_ids, " summary: ", layer.get_weights_summary())

    def get_connection_table(self):
        result = ""
        for layer in self.input_layers + [self.output_layer]:
            result += "layer.id: " + str(layer.id) + " layer outputs: " + str(layer.output_layers_ids) + " summary: " + str(layer.get_weights_summary())
        for layer in self.hidden_layers:
            result += "layer.id: " + str(layer.id) + " layer outputs: " + str(layer.output_layers_ids) + " summary: " + str(layer.get_weights_summary())
        return result

class Conv(Layer):
    def __init__(self, _id, _model, input_shape, kernel_size, depth, act_fun, _optimizer = SGDOptimizer()):
        self.id = _id
        self.model = _model
        self.act_fun = act_fun
        self.is_ending = False
        self.is_starting = False
        self.input_layers_ids = []
        self.output_layers_ids = []
        self.f_input = [] # forwarind input
        self.b_input = [] # backwarding input
        self.input_shape = input_shape
        self.input_height, self.input_width, self.input_depth = input_shape 
        self.input_flatten = int(self.input_height * self.input_width * self.input_depth)
        self.depth = int(depth) 
        self.kernel_size = int(kernel_size)
        self.output_shape = (int(self.input_height-kernel_size+1),int(self.input_width-kernel_size+1), int(depth))  
        self.output_flatten = int(self.output_shape[0] * self.output_shape[1] * self.output_shape[2])
        self.kernels_shape = (int(self.depth), int(self.input_depth), int(kernel_size), int(kernel_size)) 
        self.reshspers = {}
        self.optimizer = _optimizer
        if WEIGHT_DISTRIBUTION_MODE == DistributionMode.UNIFORM:
            # Uniform Distribution: Generate values in range (-1, 1) and shift by -0.5
            self.kernels = np.array(np.random.uniform(low=-1.0, high=1.0, size=self.kernels_shape) - 0.5)
            self.biases = np.array(np.random.uniform(low=-1.0, high=1.0, size=self.output_shape) - 0.5)
        elif WEIGHT_DISTRIBUTION_MODE == DistributionMode.NORMAL:
            # Normal Distribution: Generate values from normal distribution and shift by -0.5
            self.kernels = np.array(np.random.normal(loc=0.0, scale=1/3, size=self.kernels_shape) - 0.5)
            self.biases = np.array(np.random.normal(loc=0.0, scale=1/3, size=self.output_shape) - 0.5)
        elif WEIGHT_DISTRIBUTION_MODE == DistributionMode.GAMMA:
            # Gamma Distribution: Generate values from Gamma distribution and shift by -0.5
            self.kernels = np.array(np.random.gamma(shape=2.0, scale=1.0, size=self.kernels_shape) - 0.5)
            self.biases = np.array(np.random.gamma(shape=2.0, scale=1.0, size=self.output_shape) - 0.5)
        elif WEIGHT_DISTRIBUTION_MODE == DistributionMode.REVERSED_GAUSSIAN:
            # Reversed Gaussian: Generate values from normal distribution, shift by -0.5, and reverse by multiplying by -1
            self.kernels = np.array(get_reverse_normal_distribution(1/3, self.kernels_shape) - 0.5)
            self.biases = np.array(get_reverse_normal_distribution(1/3, self.output_shape) - 0.5)
        else:
            raise ValueError(f"Unsupported distribution mode: {WEIGHT_DISTRIBUTION_MODE}")

        

    def get_output_size(self):
        return self.output_flatten
    
    def get_reshsper(self, size_from, size_to):
        if not (size_from, size_to) in self.reshspers.keys():
            self.reshspers[(size_from, size_to)] = eye_stretch(size_from, size_to)
        return self.reshspers[(size_from, size_to)]

    def forward_prop(self, X, sender_id, deepth = 0):
        #self.f_input.append(X)
        
        self.append_to_f_input(X, sender_id)
        if any(x is None for x in self.f_input):
                return None
        
        #if len(self.f_input) < len(self.input_layers_ids): 
        #    return None
        self.I = mean_n_conv(self.f_input, self.input_shape)
        self.Z = np.zeros((self.I.shape[0], self.output_shape[0], self.output_shape[1], self.output_shape[2]))
        for img_id in range(0, self.I.shape[0]):
            for i in range(self.depth): 
                for j in range(self.input_depth): 
                    if IS_CUPY:
                        temp1 = correlate(self.I[img_id,:,:,j], self.kernels[i,j])
                        self.Z[img_id,:,:,i] += np.resize(temp1, self.Z[img_id,:,:,i].shape)
                    else:
                        self.Z[img_id,:,:,i] += correlate2d(self.I[img_id,:,:,j], self.kernels[i,j], "valid") 
                self.Z[img_id,:,:,i] += self.biases[:,:,i]
        self.A = self.act_fun.exe(self.Z)
        for layer_id in self.output_layers_ids:
            layer = self.model.get_layer(layer_id)
            if type(layer) == Conv:
                new_input = Resize(self.A.copy(), layer.input_shape)
            elif type(layer) == Layer:
                new_input = Reshape_forward_prop(self.A.copy(), layer.input_size, get_reshsper(self.output_flatten, layer.input_size))         
            else:
                raise ValueError(f"Unsupported layer type: {type(layer)}")
            
            if layer.should_thread_forward():
                thread = threading.Thread(
                    target=lambda: layer.forward_prop(new_input.copy(), self.id, deepth + 1),
                )
                thread.start()
                self.model.forward_threads.append(thread)
            else:
                #print(f"No available threads, continuing in the current thread: {threading.current_thread().name}")
                layer.forward_prop(new_input.copy(), self.id, deepth + 1)
        self.f_input = []

    def back_prop(self, E, m, alpha):
        if len(E.shape) <= 2:
            E = Reshape_back_prop(E, self.output_shape, get_reshsper(E[:, 0].shape[0], self.output_flatten))
        else:
            E = Resize(E, self.output_shape)
        self.b_input.append(E)
        if len(self.b_input) < len(self.output_layers_ids): return None
        self.E =  clip(mean_n_conv(self.b_input, self.input_shape), -error_clip_range, error_clip_range)
        dZ = self.E * self.act_fun.der(self.Z)
        self.error = np.zeros((dZ.shape[1], dZ.shape[2], dZ.shape[3]))
        self.kernels_gradient = np.zeros(self.kernels_shape)
        self.input_gradient = np.zeros((self.I.shape[0], self.input_shape[0], self.input_shape[1], self.input_shape[2]))
        for img_id in range(0, self.I.shape[0]):
            for i in range(self.depth):
                for j in range(self.input_depth):
                    if IS_CUPY:
                        temp1 = correlate(self.I[img_id,:,:,j], dZ[img_id,:,:,i])
                        self.kernels_gradient[i,j] += np.resize(temp1, self.kernels_gradient[i,j].shape)
                        temp2 = convolve(dZ[img_id, :, :, i], self.kernels[i,j])
                        self.input_gradient[img_id, :, :, j] += np.resize(temp2, self.input_gradient[img_id, :, :, j].shape)
                    else:
                        self.kernels_gradient[i,j] += correlate2d(self.I[img_id,:,:,j], dZ[img_id,:,:,i], "valid")
                        self.input_gradient[img_id,:,:,j] += convolve2d(dZ[img_id,:,:,i], self.kernels[i,j], "full")
            self.error += dZ[img_id,:,:,:]
        self.kernels_gradient[i,j] /= self.I.shape[0]
        self.input_gradient[img_id,:,:,j] /= self.I.shape[0]
        self.error /= self.I.shape[0]
        for layer_id in self.input_layers_ids:
            layer = self.model.get_layer(layer_id)
            if layer.should_thread_backward():
                thread = threading.Thread(
                    target=lambda: layer.back_prop(self.input_gradient.copy(), m, alpha),
                )
                thread.start()
                self.model.bacward_threads.append(thread)
            else:
                #print(f"No available threads, continuing in the current thread: {threading.current_thread().name} count: {threading.active_count()}")
                layer.back_prop(self.input_gradient, m, alpha)
        self.update_params(alpha)
        self.b_input = []
        #if self.is_starting:
        #    self.done_event.set()
            
    def update_params(self, alpha):
        self.kernels, self.biases = self.optimizer.update(self.kernels, self.kernels_gradient, self.biases, self.error, alpha)
            
    def deepcopy(self):
        copy = Conv(self.id, None, self.input_shape, self.kernel_size, self.depth, self.act_fun, self.optimizer.getConv())
        if self.is_ending:
            copy.set_as_ending()
        if self.is_starting:
            copy.set_as_starting()
        copy.input_layers_ids = self.input_layers_ids.copy()
        copy.output_layers_ids = self.output_layers_ids.copy()
        copy.f_input = self.f_input.copy()
        copy.b_input = self.b_input.copy()
        copy.input_shape = self.input_shape
        copy.input_depth = self.input_depth
        copy.input_height = self.input_height
        copy.input_width = self.input_width
        copy.depth = self.depth
        copy.output_shape = self.output_shape
        copy.kernels_shape = self.kernels_shape
        copy.kernels = self.kernels.copy()
        copy.biases = self.biases.copy()
        return copy
    
    def get_weights_summary(self):
        return "k: "+str(np.mean(self.kernels))+ " B: "+ str(np.mean(self.biases))
    
    def get_paint_label(self):
        return str(self.id) + "[" +str(self.input_shape)+","+str(self.output_shape)+"]"

    def __str__(self):
        return "[<layer: "+ str(self.id)+ " id: " + str(id(self)) + " model id: "+ str(id(self.model))+" in conn: "+ str(len(self.input_layers_ids)) +" out conn: "+ str(len(self.output_layers_ids))+ ">]"

    
def Resize(x, shape):
    if x[0].shape == shape:
        return x
    x_resized = np.zeros((x.shape[0], shape[0], shape[1], shape[2]))
    for i in range(0, x.shape[0]):
        x_resize_tmp = strech(x[i], (shape[0], shape[1]))
        index = min([x_resize_tmp.shape[2], shape[2]])
        x_resized[i, :, :, 0:index] = strech(x[i], (shape[0], shape[1]))[:, :, 0:index]
    return x_resized



class Storage:
    
    def __init__(self):
        pass
    

    def loadModel(path):
        with open(path, 'r') as json_file:
            model_dict = json.load(json_file)
        return Storage.dictToModel(model_dict)

    def saveModel(model, path):
        with open(path, 'w') as json_file:
            json.dump(Storage.modelToDict(model), json_file, indent=4)
        pass
    
    def modelToDict(model):
        dict_main = {}
        dict_main['settings'] = {}
        dict_main['settings']['batch_size'] = model.batch_size
        dict_main['settings']['loss_function'] = str(model.loss_function.__name__)
        dict_main['settings']['output_size'] = model.output_size
        dict_main['settings']['hidden_size'] = model.hidden_size
        dict_main['settings']['avaible_id'] = model.avaible_id
        dict_main['settings']['activation_fun'] = str(model.activation_fun.__name__)
        dict_main['settings']['convolution'] = model.convolution
        dict_main['settings']['input_size'] = model.input_size
        dict_main['settings']['input_shape'] = model.input_shape
        dict_main['settings']['kernel_size'] = model.kernel_size
        dict_main['settings']['depth'] = model.depth
        dict_main['settings']['optimizer'] = OptimizerFactory.ToDict(model.optimizer)
        dict_main['hidden_layers'] = {}
        dict_main['input_layers'] = {}
        dict_main['output_layer'] = {}
        for layer in model.hidden_layers:
            dict_main['hidden_layers'][layer.id] = Storage.layerToDict(layer)
        for layer in model.input_layers:
            dict_main['input_layers'][layer.id] = Storage.layerToDict(layer)
        dict_main['output_layer'][layer.id] = Storage.layerToDict(model.output_layer)
        return dict_main

    def layerToDict(layer):
        dict_main = {}
        dict_main['id'] = str(layer.id)
        dict_main['act_fun'] = str(layer.act_fun.__name__)
        dict_main['is_ending'] = layer.is_ending
        dict_main['is_starting'] = layer.is_starting
        dict_main['input_layers_ids'] = [str(i) for i in layer.input_layers_ids] 
        dict_main['output_layers_ids'] = [str(i) for i in layer.output_layers_ids]
        dict_main['optimizer'] = OptimizerFactory.ToDict(layer.optimizer)
        dict_reshapers = {}
        for resheper_key in layer.reshspers.keys():
            reshsper_id = str(resheper_key)
            (resheper_from, resharper_to) = resheper_key
            dict_reshapers[reshsper_id] = {}
            dict_reshapers[reshsper_id]['from'] = resheper_from
            dict_reshapers[reshsper_id]['to'] = resharper_to
            dict_reshapers[reshsper_id]["matrix"] =  get_numpy_array(layer.reshspers[resheper_key]).tolist()
        dict_main['reshapers'] = dict_reshapers
        dict_main['weights'] = {}
        if type(layer) == Layer:
            dict_main['neurons'] = layer.neurons
            dict_main['input_size'] = layer.input_size
            dict_main['weights']['W'] = get_numpy_array(layer.W).tolist()
            dict_main['weights']['B'] = get_numpy_array(layer.B).tolist()
        if type(layer) == Conv:
            dict_main['weights']['kernels'] = get_numpy_array(layer.kernels).tolist()
            dict_main['weights']['biases'] = get_numpy_array(layer.biases).tolist()
            dict_main['conv'] = {}
            dict_main['conv']['input_shape'] = layer.input_shape
            dict_main['conv']['input_height'] = layer.input_height
            dict_main['conv']['input_width'] = layer.input_width
            dict_main['conv']['input_depth'] = layer.input_depth
            dict_main['conv']['input_flatten'] = layer.input_flatten
            dict_main['conv']['depth'] = layer.depth
            dict_main['conv']['kernel_size'] = layer.kernel_size
            dict_main['conv']['output_shape'] = layer.output_shape
            dict_main['conv']['output_flatten'] = layer.output_flatten
            dict_main['conv']['kernels_shape'] = layer.kernels_shape
        return dict_main
        

    def dictToModel(dict_main):
        batch_size = dict_main['settings']['batch_size']
        loss_function_name = dict_main['settings']['loss_function']
        output_size = dict_main['settings']['output_size']
        hidden_size = dict_main['settings']['hidden_size']
        avaible_id = dict_main['settings']['avaible_id']
        activation_fun_name = dict_main['settings']['activation_fun']
        convolution = dict_main['settings']['convolution']
        input_shape = dict_main['settings']['input_shape']
        kernel_size = dict_main['settings']['kernel_size']
        depth = dict_main['settings']['depth']
        input_size = dict_main['settings']['input_size']
        loss_function = Loss.getByName(loss_function_name)
        activation_fun = Activations.getByName(activation_fun_name)
        optimizer = OptimizerFactory.FromDict(dict_main['settings']['optimizer'])
        model = Model(input_size, hidden_size, output_size, loss_function, activation_fun, 1, optimizer)
        model.batch_size = batch_size
        model.avaible_id = avaible_id
        model.convolution = convolution
        model.kernel_size = kernel_size
        model.depth = depth
        model.input_layers = []
        model.hidden_layers = []

        for layer_id, layer_dict in dict_main['input_layers'].items():
            layer = Storage.dictToLayer(layer_dict, model)
            model.input_layers.append(layer)

        for layer_id, layer_dict in dict_main['hidden_layers'].items():
            layer = Storage.dictToLayer(layer_dict, model)
            model.hidden_layers.append(layer)

        for layer_id, layer_dict in dict_main['output_layer'].items():
            layer = Storage.dictToLayer(layer_dict, model)
            model.output_layer = layer
        return model


    def dictToLayer(layer_dict, model):
        layer_id = str(layer_dict['id'])
        act_fun_name = layer_dict['act_fun']
        is_ending = layer_dict['is_ending']
        is_starting = layer_dict['is_starting']
        input_layers_ids = layer_dict['input_layers_ids']
        output_layers_ids = layer_dict['output_layers_ids']
        reshapers = layer_dict['reshapers']
        act_fun = Activations.getByName(act_fun_name)
        optimizer = OptimizerFactory.FromDict(layer_dict['optimizer'])
        if 'conv' in layer_dict:
            conv_dict = layer_dict['conv']
            layer = Conv(layer_id, model, conv_dict['input_shape'], conv_dict['kernel_size'], conv_dict['depth'], act_fun, optimizer)
            layer.input_height = conv_dict['input_height']
            layer.input_width = conv_dict['input_width']
            layer.input_depth = conv_dict['input_depth']
            layer.input_flatten = conv_dict['input_flatten']
            layer.output_shape = conv_dict['output_shape']
            layer.output_flatten = conv_dict['output_flatten']
            layer.kernels_shape = conv_dict['kernels_shape']
        else:
            input_size = layer_dict['input_size']
            neurons = layer_dict['neurons']
            layer = Layer(layer_id, model, input_size, neurons, act_fun, Layer_Type.RANDOM,  optimizer.getDense())
        
        if is_ending: layer.set_as_ending()
        if is_starting : layer.set_as_starting()
        layer.input_layers_ids = input_layers_ids
        layer.output_layers_ids = output_layers_ids
        reshapers_dict = {}
        for reshaper_id, reshaper_info in reshapers.items():
            reshaper_from = reshaper_info['from']
            reshaper_to = reshaper_info['to']
            reshaper_matrix = np.array(reshaper_info['matrix'])
            reshapers_dict[(reshaper_from, reshaper_to)] = reshaper_matrix
        layer.reshspers = reshapers_dict

        if 'weights' in layer_dict:
            weights_dict = layer_dict['weights']
            if 'W' in weights_dict:
                layer.W = np.array(weights_dict['W'])
            if 'B' in weights_dict:
                layer.B = np.array(weights_dict['B'])
            if 'kernels' in weights_dict:
                layer.kernels = np.array(weights_dict['kernels'])
            if 'biases' in weights_dict:
                layer.biases = np.array(weights_dict['biases'])
        return layer



import time
import random
from ..action import Action
from ..structure import *

class QLearningAgent:
    def __init__(self, learning_rate=0.1, discount_factor=0.9, exploration_prob=0.3):
        self.learning_rate = learning_rate
        self.discount_factor = discount_factor
        self.exploration_prob = exploration_prob
        self.q_table = {}  # Słownik przechowujący wartości Q dla każdej pary (stan, akcja)

    def get_q_value(self, state, action):
        return self.q_table.get((state, action), 0.0)

    def update_q_value(self, state, action, new_value):
        self.q_table[(state, action)] = new_value

    def choose_action(self, actions, state):
        if random.uniform(0, 1) < self.exploration_prob:
            return random.choice(actions)
        else:
            q_values = [self.get_q_value(state, action) for action in actions]
            max_q = max(q_values)
            best_actions = [action for action, q_value in zip(actions, q_values) if q_value == max_q]
            return random.choice(best_actions)

# Globalny agent Q-learninga
global_agent = QLearningAgent()

async def get_action(M, max_time_for_dec, epochs, X_train, Y_train, simulation_score):
    size_of_changes = len(Action.generate_all_actions(M))
    if size_of_changes == 0:
        print("Error")
        return None, 0

    deadline = time.time() + max_time_for_dec
    deepth = 0
    rollouts = 0

    while time.time() < deadline or rollouts <= size_of_changes:
        # Rozpoczęcie nowej symulacji - agent nie zna stanu ani akcji
        state = M.get_state_representation() if M else None
        actions = Action.generate_all_actions(M)

        # Wybór akcji przez agenta
        action = global_agent.choose_action(actions, state)

        # Wykonanie akcji na kopii obiektu M, jeśli istnieje
        new_M = M.deepcopy() if M else None
        if new_M:
            action.execute(new_M)
            #new_M.apply_action(action)

        # Uzyskanie nagrody za wykonanie akcji
        reward = simulation_score.scoreFun(new_M, epochs, X_train, Y_train) if new_M else 0

        # Aktualizacja wartości Q
        if state and action:
            next_state = new_M.get_state_representation()
            q_value = global_agent.get_q_value(state, action)
            next_max_q_value = max(global_agent.get_q_value(next_state, next_action) for next_action in actions)
            new_q_value = q_value + global_agent.learning_rate * (reward + global_agent.discount_factor * next_max_q_value - q_value)
            global_agent.update_q_value(state, action, new_q_value)

        rollouts += 1

    if time.time() > deadline:
        print("More time was needed to analyze all possibilities at least once")

    # Ostateczne wybranie najlepszej akcji po zakończeniu symulacji
    best_action = global_agent.choose_action(actions, state)

    return best_action, deepth, rollouts

# def scoreFun(M, epochs, X_train, Y_train, simulation_score):
#     acc, history = M.gradient_descent(X_train, Y_train, epochs, LearningRateScheduler(LearningRateScheduler.CONSTANT, 0.1) , True)
#     return simulation_score.grade(acc, history)
#     #return max(1.e-17, max_loss - history.get_last('loss'))



import os
from enum import Enum

class DistributionMode(Enum):
    UNIFORM = 'uniform'
    NORMAL = 'normal'
    GAMMA = 'gamma'
    REVERSED_GAUSSIAN = 'reversed_gaussian'
    
WEIGHT_DISTRIBUTION_MODE = DistributionMode.NORMAL
error_clip_range = 600 # In back propagation we are clipping the error range to prevent exploding gradients
WEIGHTS_CLIP_RANGE = 3 # After weight update we are clipping all weights to prevent exploding gradients
weights_clip_range = 400 # After weight update we are clipping all weights to prevent exploding gradients
LARGE_MAX = 2**128
MAX_THREADS = 1#max(1, int(os.cpu_count() * 0.5))
VERSION = 'R3'




import sys
sys.path.append('.')
sys.path.append('../')
import unittest
import numpy as np
import growingnn as gnn
from growingnn.structure import SimulationScheduler
from testSuite import mode

class TestModelTraining(unittest.TestCase):

    def setUp(self):
        global mode
        if mode == 'cpu':
            gnn.switch_to_cpu()
        elif mode == 'gpu':
            gnn.switch_to_gpu()

        self.datasize = 20
        self.datadimensionality = 10
        self.classes = 3
        self.x_train = np.random.random((self.datadimensionality, self.datasize))
        self.y_train = np.random.randint(self.classes, size=(self.datasize, ))
        self.x_test = np.random.random((self.datadimensionality, int(self.datasize / 2)))
        self.y_test = np.random.randint(self.classes, size=(int(self.datasize / 2), ))

        self.x_conv_train = np.random.random((self.datasize, self.datadimensionality, self.datadimensionality, 1))
        self.y_conv_train = np.random.randint(self.classes, size=(self.datasize, ))
        self.x_conv_test = np.random.random((int(self.datasize / 2), self.datadimensionality, self.datadimensionality, 1))
        self.y_conv_test = np.random.randint(self.classes, size=(int(self.datasize / 2), ))
        self.labels = range(0, self.classes)
        self.optimizer = gnn.SGDOptimizer()
        self.simulation_time = 2
        self.simulation_alg = gnn.montecarlo_alg
        print("self.x_train: ", self.x_train.shape)
        print("self.y_train: ", self.y_train.shape)

    # def test_train_dense_GPU(self):
    #     mode = getattr(self, 'mode', 'cpu')  # Default to 'cpu' if 'mode' is not set
    #     if mode == 'cpu':
    #         return
    #     # Wykonywanie treningu modelu z małym zbiorem danych
    #     gnn.switch_to_gpu()
    #     try:
    #         self.train_dense()
    #     except Exception as e:
    #         self.fail(f"Model training failed with exception: {e}")

    def test_train_dense_CPU(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.train_dense()
        try:
            self.train_dense()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}")

    # def test_train_conv_GPU(self):
    #     mode = getattr(self, 'mode', 'cpu')  # Default to 'cpu' if 'mode' is not set
    #     if mode == 'cpu':
    #         return
    #     # Wykonywanie treningu modelu z małym zbiorem danych
    #     gnn.switch_to_gpu()
    #     try:
    #         self.train_conv()
    #     except Exception as e:
    #         self.fail(f"Model training failed with exception: {e}")

    def test_train_conv_CPU(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.train_conv()
        try:
            self.train_conv()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}")

    def test_train_dense_CPU_SGD(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.SGDOptimizer()
        self.train_dense()
        try:
            self.train_dense()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}")

    def test_train_conv_CPU_Adam(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.AdamOptimizer()
        self.train_conv()
        try:
            self.train_conv()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}") #acc = Model.get_accuracy(Model.get_predictions(self.forward_prop(X)),Y)

    def test_train_conv_CPU_Adam_vs_SGD(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.AdamOptimizer()
        model_adam = self.train_dense()

        self.optimizer = gnn.SGDOptimizer()
        model_sgd = self.train_dense()

        acc_adam = gnn.Model.get_accuracy(gnn.Model.get_predictions(model_adam.forward_prop(self.x_train)), self.y_train)
        acc_sgd = gnn.Model.get_accuracy(gnn.Model.get_predictions(model_sgd.forward_prop(self.x_train)), self.y_train)
        
        self.assertEqual(acc_adam >= acc_sgd * 0.9, True, "Adam optimzier should have better result than SGD" + str(acc_adam) + " > " + str(acc_sgd))
        try:
            self.train_conv()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}") #acc = Model.get_accuracy(Model.get_predictions(self.forward_prop(X)),Y)

    def test_train_dense_CPU_SGD_SIMULATION(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.SGDOptimizer()
        self.simulation_time = 30
        self.train_dense()
        try:
            self.train_dense()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}")

    def test_train_conv_CPU_Adam_SIMULATION(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.AdamOptimizer()
        self.simulation_time = 30
        self.train_conv()
        try:
            self.train_conv()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}") #acc = Model.get_accuracy(Model.get_predictions(self.forward_prop(X)),Y)

    def test_train_dense_CPU_SGD_monte_carlo(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.SGDOptimizer()
        self.simulation_time = 10
        self.simulation_alg = gnn.montecarlo_alg
        self.train_dense()
        try:
            self.train_dense()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}")

    def test_train_dense_CPU_SGD_greedy(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.SGDOptimizer()
        self.simulation_time = 10
        self.simulation_alg = gnn.greedy_alg
        self.train_dense()
        try:
            self.train_dense()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}")
            
    def test_train_dense_CPU_SGD_random(self):
        # Wykonywanie treningu modelu z małym zbiorem danych
        gnn.switch_to_cpu()
        self.optimizer = gnn.SGDOptimizer()
        self.simulation_time = 10
        self.simulation_alg = gnn.random_alg
        self.train_dense()
        try:
            self.train_dense()
        except Exception as e:
            self.fail(f"Model training failed with exception: {e}")
            
    def train_conv(self):
        return gnn.trainer.train(
            x_train=self.x_conv_train,
            y_train=self.y_conv_train,
            x_test=self.x_conv_test,
            y_test=self.y_conv_test,
            labels=self.labels,
            input_paths=1,
            path="./result",
            model_name="GNN_model",
            epochs=3,
            generations=3,
            input_size=self.datadimensionality,
            hidden_size=self.datadimensionality,
            output_size=self.classes,
            input_shape=(self.datadimensionality, self.datadimensionality, 1),
            kernel_size=2,
            batch_size=1,
            simulation_scheduler = SimulationScheduler(SimulationScheduler.PROGRESS_CHECK, simulation_time = self.simulation_time, simulation_epochs = 2), 
            deepth=2,
            simulation_alg=self.simulation_alg,
            optimizer=self.optimizer
        )

    def train_dense(self):
        return gnn.trainer.train(
            x_train=self.x_train,
            y_train=self.y_train,
            x_test=self.x_test,
            y_test=self.y_test,
            labels=self.labels,
            input_paths=1,
            path="./result",
            model_name="GNN_model",
            epochs=3,
            generations=3,
            input_size=self.datadimensionality,
            hidden_size=self.datadimensionality,
            output_size=self.classes,
            input_shape=None,
            kernel_size=None,
            batch_size=1,
            simulation_scheduler = SimulationScheduler(SimulationScheduler.PROGRESS_CHECK, simulation_time = self.simulation_time, simulation_epochs = 2), 
            deepth=None,
            simulation_alg=self.simulation_alg,
            optimizer=self.optimizer
        )


if __name__ == '__main__':
    unittest.main()



from numba import jit
import numpy as np

from growingnn.config import WEIGHTS_CLIP_RANGE

class Optimizer:
    def __init__(self, weights_clip_range=WEIGHTS_CLIP_RANGE):
        self.weights_clip_range = weights_clip_range
        self.updateType("SGD", "Dense")

    def updateType(self, _optimzierType, _layerType):
        self.optimzierType = _optimzierType
        self.layerType = _layerType
        pass

    #@staticmethod
    #@jit(nopython=True)
    def clip_and_fix(params, clip_range):
        params = np.clip(params, -clip_range, clip_range)
        params = np.nan_to_num(params, nan=np.nanmean(params))
        #params[params == -np.inf] = -1.0
        #params[params == np.inf] = 1.0
        return params

    def ToDict(self):
        dict = {}
        dict["optimzierType"] = self.optimzierType
        dict["layerType"] = self.layerType
        dict["weights_clip_range"] = self.weights_clip_range
        return dict
    
    def FromDict(dict):
        return Optimizer(dict["weights_clip_range"])
    
    def update(self, params, grads, alpha):
        raise NotImplementedError("This method should be implemented by subclasses.")

class SGDOptimizer(Optimizer):

    def __init__(self, weights_clip_range=WEIGHTS_CLIP_RANGE):
        super().__init__(weights_clip_range)
        self.updateType("SGD", "Dense")
        
    @staticmethod
    @jit(nopython=True)
    def sgd_update(params, grads, alpha):
         return params.astype(np.float64) - alpha * grads

    def ToDict(self):
        dict = {}
        dict["optimzierType"] = self.optimzierType
        dict["layerType"] = self.layerType
        dict["weights_clip_range"] = self.weights_clip_range
        return dict
    
    def FromDict(dict):
        return SGDOptimizer(dict["weights_clip_range"])
    
    def update(self, params, grads, alpha):
        return SGDOptimizer.clip_and_fix(SGDOptimizer.sgd_update(params, grads, alpha), 
                                         WEIGHTS_CLIP_RANGE)
    
    def getDense(self):
        return self

    def getConv(self):
        return ConvSGDOptimizer(self.weights_clip_range)

class AdamOptimizer(Optimizer):
    def __init__(self, beta1=0.9, beta2=0.999, epsilon=1e-8, weights_clip_range=WEIGHTS_CLIP_RANGE):
        super().__init__(weights_clip_range)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.m = None
        self.v = None
        self.t = 0
        self.updateType("Adam", "Dense")

    def ToDict(self):
        dict = {}
        dict["optimzierType"] = self.optimzierType
        dict["layerType"] = self.layerType
        dict["beta1"] = self.beta1
        dict["beta2"] = self.beta2
        dict["epsilon"] = self.epsilon
        dict["weights_clip_range"] = self.weights_clip_range
        return dict
    
    def FromDict(dict):
        return AdamOptimizer(dict["beta1"], dict["beta2"], dict["epsilon"], dict["weights_clip_range"])
    
    @staticmethod
    @jit(nopython=True)
    def adam_update(params, grads, m, v, t, alpha, beta1, beta2, epsilon):
        m = np.copy(beta1 * m + (1 - beta1) * grads)
        v = np.copy(beta2 * v + (1 - beta2) * (grads ** 2))
        m_hat = m / (1 - beta1 ** t)
        v_hat = v / (1 - beta2 ** t)
        params = params - alpha * m_hat / (np.sqrt(v_hat) + epsilon)
        return params, m, v

    def update(self, params, grads, alpha):
        if self.m is None or self.v is None or self.m.shape != params.shape:
            reset = True
        else:
            reset = False
        if reset:
            self.m = np.zeros_like(params)
            self.v = np.zeros_like(params)
        self.t += 1
        params, self.m, self.v = self.adam_update(
            params, grads, self.m, self.v, self.t, alpha, self.beta1, self.beta2, self.epsilon
        )
        return AdamOptimizer.clip_and_fix(params, self.weights_clip_range)

    def getDense(self):
        return self
    
    def getConv(self):
        return ConvAdamOptimizer(self.beta1, self.beta2, self.epsilon, self.weights_clip_range)
    
class ConvSGDOptimizer(Optimizer):

    def __init__(self, weights_clip_range=WEIGHTS_CLIP_RANGE):
        super().__init__(weights_clip_range)
        self.updateType("SGD", "Conv")

    def ToDict(self):
        dict = {}
        dict["optimzierType"] = self.optimzierType
        dict["layerType"] = self.layerType
        dict["weights_clip_range"] = self.weights_clip_range
        return dict
    
    def FromDict(dict):
        return ConvSGDOptimizer(dict["weights_clip_range"])
    
    @staticmethod
    @jit(nopython=True)
    def conv_sgd_update(kernels, kernel_grads, biases, bias_grads, alpha):
        kernels = kernels - alpha * kernel_grads
        biases = biases - alpha * bias_grads
        return kernels, biases

    def update(self, kernels, kernel_grads, biases, bias_grads, alpha):
        kernels, biases = self.conv_sgd_update(kernels, kernel_grads, biases, bias_grads, alpha)
        kernels = ConvSGDOptimizer.clip_and_fix(kernels, self.weights_clip_range)
        biases = ConvSGDOptimizer.clip_and_fix(biases, self.weights_clip_range)
        return kernels, biases
    
    def getDense(self):
        return SGDOptimizer(self.weights_clip_range)
    
    def getConv(self):
        return self
    
class ConvAdamOptimizer(Optimizer):
    def __init__(self, beta1=0.9, beta2=0.999, epsilon=1e-8, weights_clip_range=WEIGHTS_CLIP_RANGE):
        super().__init__(weights_clip_range)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.m_kernels = None
        self.v_kernels = None
        self.m_biases = None
        self.v_biases = None
        self.t = 0
        self.updateType("Adam", "Conv")

    def ToDict(self):
        dict = {}
        dict["optimzierType"] = self.optimzierType
        dict["layerType"] = self.layerType
        dict["beta1"] = self.beta1
        dict["beta2"] = self.beta2
        dict["epsilon"] = self.epsilon
        dict["weights_clip_range"] = self.weights_clip_range
        return dict
    
    def FromDict(dict):
        return ConvAdamOptimizer(dict["beta1"], dict["beta2"], dict["epsilon"], dict["weights_clip_range"])
    
    @staticmethod
    @jit(nopython=True)
    def conv_adam_update(kernels, kernel_grads, biases, bias_grads, m_kernels, v_kernels, m_biases, v_biases, t, alpha, beta1, beta2, epsilon):
        # Adam for kernels
        m_kernels = beta1 * m_kernels + (1 - beta1) * kernel_grads
        v_kernels = beta2 * v_kernels + (1 - beta2) * (kernel_grads ** 2)
        m_hat_kernels = m_kernels / (1 - beta1 ** t)
        v_hat_kernels = v_kernels / (1 - beta2 ** t)
        kernels = kernels - alpha * m_hat_kernels / (np.sqrt(v_hat_kernels) + epsilon)
        
        # Adam for biases
        m_biases = beta1 * m_biases + (1 - beta1) * bias_grads
        v_biases = beta2 * v_biases + (1 - beta2) * (bias_grads ** 2)
        m_hat_biases = m_biases / (1 - beta1 ** t)
        v_hat_biases = v_biases / (1 - beta2 ** t)
        biases = biases - alpha * m_hat_biases / (np.sqrt(v_hat_biases) + epsilon)
        
        return kernels, biases, m_kernels, v_kernels, m_biases, v_biases

    def update(self, kernels, kernel_grads, biases, bias_grads, alpha):
        if self.m_kernels is None:
            self.m_kernels = np.zeros_like(kernels)
        if self.v_kernels is None:
            self.v_kernels = np.zeros_like(kernels)
        if self.m_biases is None:
            self.m_biases = np.zeros_like(biases)
        if self.v_biases is None:
            self.v_biases = np.zeros_like(biases)

        self.t += 1
        kernels, biases, self.m_kernels, self.v_kernels, self.m_biases, self.v_biases = self.conv_adam_update(
            kernels, kernel_grads, biases, bias_grads, self.m_kernels, self.v_kernels, self.m_biases, self.v_biases,
            self.t, alpha, self.beta1, self.beta2, self.epsilon
        )
        kernels = ConvAdamOptimizer.clip_and_fix(kernels, self.weights_clip_range)
        biases = ConvAdamOptimizer.clip_and_fix(biases, self.weights_clip_range)
        return kernels, biases
    
    def getDense(self):
        return AdamOptimizer(self.beta1, self.beta2, self.epsilon, self.weights_clip_range)
    
    def getConv(self):
        return self
    
class OptimizerFactory:
    SGD = "SGD"
    Adam = "Adam"

    def __init__(self, weights_clip_range=WEIGHTS_CLIP_RANGE, **kwargs):
        self.weights_clip_range = weights_clip_range
        self.kwargs = kwargs

    def create_optimizer(self, optimizer_type, layer_type):
        if layer_type == "Dense":
            if optimizer_type == OptimizerFactory.SGD:
                return SGDOptimizer(self.weights_clip_range)
            elif optimizer_type == OptimizerFactory.Adam:
                return AdamOptimizer(**self.kwargs)
        elif layer_type == "Conv":
            if optimizer_type == OptimizerFactory.SGD:
                return ConvSGDOptimizer(self.weights_clip_range)
            elif optimizer_type == OptimizerFactory.Adam:
                # Assuming ConvOptimizer can also support Adam-like optimizers, or you could implement a ConvAdamOptimizer class
                return ConvAdamOptimizer(self.weights_clip_range)  # or ConvAdamOptimizer(self.alpha, **self.kwargs)
        else:
            raise ValueError("Unsupported layer type or optimizer type")
        
    def copy(optimizer):
        return OptimizerFactory.FromDict(OptimizerFactory.ToDict(optimizer))
    
    def ToDict(optimizer):
        return optimizer.ToDict()
    
    def FromDict(dict):
        if dict["layerType"] ==  "Conv":
            if dict["optimzierType"] ==  "SGD":
                return ConvSGDOptimizer.FromDict(dict)
            elif dict["optimzierType"] ==  "Adam":
                return ConvAdamOptimizer.FromDict(dict)
        elif dict["layerType"] ==  "Dense":
            if dict["optimzierType"] ==  "SGD":
                return SGDOptimizer.FromDict(dict)
            elif dict["optimzierType"] ==  "Adam":
                return AdamOptimizer.FromDict(dict)    


import asyncio
import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import numpy as np
import unittest
import random
from testSuite import mode

shape = 20
epochs = 5

class TestingTrain(unittest.TestCase):

    def setUp(self):
        global mode
        if mode == 'cpu':
            gnn.switch_to_cpu()
        elif mode == 'gpu':
            gnn.switch_to_gpu()

    def test_simple_SGD_train(self):
        M = gnn.structure.Model(shape,shape,2, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1, gnn.optimizers.SGDOptimizer())
        x = np.random.rand(shape, shape)
        y = np.random.randint(2, size=(shape,))
        lr_scheduler = gnn.structure.LearningRateScheduler(gnn.structure.LearningRateScheduler.PROGRESIVE, 0.03, 0.8)
        gnn.painter.draw(M, "input_test.html")
        acc, _ = M.gradient_descent(x, y, epochs, lr_scheduler)
        self.assertEqual(acc >= 0.3, True)

    def test_simple_Adam_train(self):
        M = gnn.structure.Model(shape,shape,2, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1, gnn.optimizers.AdamOptimizer())
        x = np.random.rand(shape, shape)
        y = np.random.randint(2, size=(shape,))
        lr_scheduler = gnn.structure.LearningRateScheduler(gnn.structure.LearningRateScheduler.PROGRESIVE, 0.03, 0.8)
        gnn.painter.draw(M, "input_test.html")
        acc, _ = M.gradient_descent(x, y, epochs, lr_scheduler)
        self.assertEqual(acc >= 0.3, True)

    def test_res_SGD_structure(self):
        M = gnn.structure.Model(shape, shape, 2, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1, gnn.optimizers.SGDOptimizer())
        M.add_res_layer('init_0', 1)
        x = np.asarray(np.random.rand(shape, shape))
        for _ in range(10):
            output = M.forward_prop(x * float(np.random.ranf(1)))
        self.assertEqual(output.any(), True)

    def test_res_Adam_structure(self):
        M = gnn.structure.Model(shape, shape, 2, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1, gnn.optimizers.AdamOptimizer())
        M.add_res_layer('init_0', 1)
        x = np.asarray(np.random.rand(shape, shape))
        for _ in range(10):
            output = M.forward_prop(x * float(np.random.ranf(1)))
        self.assertEqual(output.any(), True)

    def test_conv_SGD_structure(self):
        M = gnn.structure.Model(shape, shape, shape, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1, gnn.optimizers.SGDOptimizer())
        M.set_convolution_mode((shape, shape, 1), shape, 1)
        M.add_res_layer('init_0', 1)
        x = np.random.random((shape, shape, shape, 1))
        for _ in range(10):
            output = M.forward_prop(x * float(np.random.ranf(1)))
        self.assertEqual(output.any(), True)

    def test_conv_Adam_structure(self):
        M = gnn.structure.Model(shape, shape, shape, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1, gnn.optimizers.AdamOptimizer())
        M.set_convolution_mode((shape, shape, 1), shape, 1)
        M.add_res_layer('init_0', 1)
        x = np.random.random((shape, shape, shape, 1))
        for _ in range(10):
            output = M.forward_prop(x * float(np.random.ranf(1)))
        self.assertEqual(output.any(), True)

if __name__ == '__main__':
    unittest.main()



from .painter import *
from .structure import *
from .action import *
import asyncio
from .Simulation import *
from .Simulation.ScoreFunctions import *
import os
from .helpers import convert_to_desired_type

def train(x_train, x_test, y_train, y_test, labels, path, model_name, epochs, generations, input_size, hidden_size, output_size, input_shape, kernel_size, deepth, batch_size = 128, simulation_set_size = 20, simulation_alg = montecarlo_alg, sim_set_generator = create_simulation_set_SAMLE, simulation_scheduler = SimulationScheduler(SimulationScheduler.PROGRESS_CHECK, simulation_time = 60, simulation_epochs = 20), lr_scheduler = LearningRateScheduler(LearningRateScheduler.PROGRESIVE, 0.03, 0.8), loss_function = Loss.multiclass_cross_entropy, activation_fun = Activations.Sigmoid, input_paths = 1, sample_sub_generator = None, simulation_score = Simulation_score(), optimizer = SGDOptimizer()):
    x_train = convert_to_desired_type(x_train)
    x_test = convert_to_desired_type(x_test)
    y_train = convert_to_desired_type(y_train)
    y_test = convert_to_desired_type(y_test)
    print("x_train: ", type(x_train))
    print("x_test: ", type(x_test))
    print("y_train: ", type(y_train))
    print("y_test: ", type(y_test))
    if not os.path.exists(path):
        os.mkdir(path, 0o777)
    if not os.path.exists(path): os.mkdir(path, 0o777)
    hist = {'train': [], 'test': []}
    hist_detail = History(['accuracy', 'loss'])
    model_path = path+model_name
    hist_path = path+model_name+"_hist"
    M = Model(input_size, hidden_size, output_size, loss_function, activation_fun, input_paths, optimizer)
    if input_shape != None:
            M.set_convolution_mode(input_shape, kernel_size, deepth)
    M.batch_size = batch_size
    acc = Model.get_accuracy(Model.get_predictions(M.forward_prop(x_test)),y_test)
    print("model is ready, starting accuracy: ", acc)

    sim_x, sim_y = sim_set_generator(x_train, y_train, simulation_set_size)
    draw(M, model_path+ 'cifar_init.html')
    for i in range(0,generations):
        draw(M, model_path+'_graph_'+ str(hist_detail.last_img_id)+"bef.html")
        new_acc, new_hist = M.gradient_descent(x_train, y_train, epochs, lr_scheduler, False, True, model_path+"_gen_"+str(i))
        hist_detail.merge(new_hist)
        hist_detail.append('iteration_acc_train', new_acc)
        hist_detail.append('iteration_acc_test', M.evaluate(x_test, y_test))
        if simulation_scheduler.can_simulate(i, hist_detail):
            correc_desc = "[iteration: "+str(i)+"] No correction detected acc: " + str(new_acc)+ " starting simulation." 
            hist_detail.description += correc_desc+'\n'            
            loop = asyncio.get_event_loop()
            action, deepth, rollouts = loop.run_until_complete(simulation_alg.get_action(M.deepcopy(), simulation_scheduler.simulation_time, simulation_scheduler.simulation_epochs, sim_x, sim_y, simulation_score))
            size_of_changes = len(Action.generate_all_actions(M))
            action_desc = "[iteration: "+str(i)+"] Best action found after simulation: "+ str(action)+ " deepth of tree searched: "+ str(deepth) + " number of rollouts: "+ str(rollouts) + " size_of_changes: " + str(size_of_changes)
            hist_detail.description += action_desc+"\n"
            action.execute(M)
        hist_detail.save(hist_path)
        Storage.saveModel(M, model_path + "epoch_" + str(i) + "save.json")
        draw(M, model_path+'_graph_'+ str(hist_detail.last_img_id)+".html")
        hist_detail.last_img_id += 1
        hist_detail.draw_hist(model_path+"_history_detail", ".")
        if hist_detail.get_last('iteration_acc_train') > hist_detail.best_train_acc:
            hist_detail.description += '[iteration: '+str(i)+'] Rewriting best model for train acc prev: ' + str(hist_detail.best_train_acc) + " new: " + str(hist_detail.get_last('iteration_acc_train')) + "\n"
            hist_detail.best_train_acc = hist_detail.get_last('iteration_acc_train')
            if sample_sub_generator != None:
                sample_sub_generator(M, model_path+"train_", labels, x_train, y_train, x_test, y_test)
        if hist_detail.get_last('iteration_acc_test') > hist_detail.best_test_acc:
            hist_detail.description += '[iteration: '+str(i)+'] Rewriting best model for test acc prev: ' + str(hist_detail.best_test_acc) + " new: " + str(hist_detail.get_last('iteration_acc_test')) + "\n"
            hist_detail.best_test_acc = hist_detail.get_last('iteration_acc_test')
            if sample_sub_generator != None:
                sample_sub_generator(M, model_path+"test_", labels, x_train, y_train, x_test, y_test)
    draw(M, model_path+"_graph.html")
    return M



from .structure import *
import numpy as np
import math

class Action:
    def __init__(self, _params):
        self.params = _params
        pass

    def execute(self, Model):
        pass

    def can_be_infulenced(self, by_action):
        pass

    def generate_all_actions(Model):
        result = []
        adding_layer_seq_actions = Add_Seq_Layer.generate_all_actions(Model)
        adding_layer_res_actions = Add_Res_Layer.generate_all_actions(Model)
        adding_layer_conv_seq_actions = Add_Seq_Conv_Layer.generate_all_actions(Model)
        adding_layer_conv_res_actions = Add_Res_Conv_Layer.generate_all_actions(Model)
        delete_layer_actions = Del_Layer.generate_all_actions(Model)
        #empty_action = Empty.generate_all_actions(Model)
        
        result += adding_layer_seq_actions
        result += adding_layer_res_actions
        result += adding_layer_conv_seq_actions
        result += adding_layer_conv_res_actions
        result += delete_layer_actions
        #result += empty_action
        return result

class Add_Seq_Layer(Action):
    def execute(self, Model):
        Model.add_norm_layer(self.params[0], self.params[1], self.params[2])
    
    def can_be_infulenced(self, by_action):
        if type(by_action) is Del_Layer:
            if by_action.params == self.params[0]: return True
            if by_action.params == self.params[1]: return True
        return False

    def generate_all_actions(Model):
        pairs = Model.get_sequence_connection()
        pairs = delete_repetitions(pairs)
        actions = []
        for pair in pairs:
            if type(Model.get_layer(pair[1])) != Conv:
                actions.append(Add_Seq_Layer([pair[0], pair[1], Layer_Type.EYE]))
        return actions
    
    def __str__(self):
        return " ( Add Seq Layer Action: " + str(self.params) + " ) "

class Empty(Action):
    def execute(self, Model):
        pass

    def can_be_infulenced(self, by_action):
        return False

    def generate_all_actions(Model):
        return [Empty(None)]

    def __str__(self):
        return " ( Empty Action )"
    
class Add_Res_Layer(Action):
    def execute(self, Model):
        Model.add_res_layer(self.params[0], self.params[1], self.params[2])
    
    def can_be_infulenced(self, by_action):
        if type(by_action) is Del_Layer:
            if by_action.params == self.params[0]: return True
            if by_action.params == self.params[1]: return True
        return False

    def generate_all_actions(Model):
        pairs = Model.get_all_childrens_connections()
        pairs = delete_repetitions(pairs)
        actions = []
        for pair in pairs:
            if type(Model.get_layer(pair[1])) != Conv:
                actions.append(Add_Res_Layer([pair[0], pair[1], Layer_Type.ZERO]))
        for pair in pairs:
            if type(Model.get_layer(pair[1])) != Conv:
                actions.append(Add_Res_Layer([pair[0], pair[1], Layer_Type.RANDOM]))
        for pair in pairs:
            if type(Model.get_layer(pair[1])) != Conv:
                actions.append(Add_Res_Layer([pair[0], pair[1], Layer_Type.EYE]))
        return actions
    
    def __str__(self):
        return " ( Add Res Layer Action: " + str(self.params) + " ) "

class Del_Layer(Action):
    def execute(self, Model):
        #print("Add_Layer: ", self.params)
        Model.remove_layer(self.params)

    def can_be_infulenced(self, by_action):
        return False

    def generate_all_actions(Model):
        actions = []
        for layer_hidden in Model.hidden_layers:
            actions.append(Del_Layer(layer_hidden.id))
        return actions

    def __str__(self):
        return " ( Del Layer Action: " + str(self.params) + " ) "
    

class Add_Seq_Conv_Layer(Action):
    def execute(self, Model):
        Model.add_conv_norm_layer(self.params[0], self.params[1])

    def can_be_infulenced(self, by_action):
        if type(by_action) is Del_Layer:
            if by_action.params == self.params[0]: return True
            if by_action.params == self.params[1]: return True
        return False

    def generate_all_actions(Model):
        pairs = Model.get_sequence_connection()
        pairs = delete_repetitions(pairs)
        actions = []
        for pair in pairs:
            layer_from = Model.get_layer(pair[0])
            layer_to = Model.get_layer(pair[1])
            #print("type(layer_from): ", type(layer_from))
            if type(layer_from) == Conv:
                if type(Model.get_layer(pair[1])) == Conv:
                    i = layer_from.output_shape[0]
                    d = np.clip(layer_to.depth, 1, 10)
                    k = np.clip(layer_to.kernel_size, 1, i)  
                    output_shape = (i-k+1,i-k+1, d)  
                    output_flatten = output_shape[0] * output_shape[1] * output_shape[2]
                    #if output_shape == layer_to.input_shape:
                    actions.append(Add_Seq_Conv_Layer(pair))                
                elif type(layer_to) == Layer:
                    o = layer_to.input_size
                    i = layer_from.output_shape[0]
                    d = np.clip(math.floor(o ** ( 1 / 3)), 1, 10)
                    k = np.clip(i + 1 - math.ceil(o ** ( 1 / 3)), 1, i)
                    output_shape = (i-k+1,i-k+1, d)  
                    output_flatten = output_shape[0] * output_shape[1] * output_shape[2]
                    #if output_flatten == layer_to.input_size:
                    actions.append(Add_Seq_Conv_Layer(pair))
        return actions
    
    def __str__(self):
        return " ( Add Seq Conv Layer Action: " + str(self.params) + " ) "
    

class Add_Res_Conv_Layer(Action):
    def execute(self, Model):
        Model.add_conv_res_layer(self.params[0], self.params[1])
    
    def can_be_infulenced(self, by_action):
        if type(by_action) is Del_Layer:
            if by_action.params == self.params[0]: return True
            if by_action.params == self.params[1]: return True
        return False

    def generate_all_actions(Model):
        pairs = Model.get_all_childrens_connections()
        pairs = delete_repetitions(pairs)
        actions = []
        for pair in pairs:
            layer_from = Model.get_layer(pair[0])
            layer_to = Model.get_layer(pair[1])
            if type(layer_from) == Conv:
                if type(Model.get_layer(pair[1])) == Conv:
                    i = layer_from.output_shape[0]
                    d = np.clip(layer_to.depth, 1, 10)
                    k = np.clip(layer_to.kernel_size, 1, i)  
                    output_shape = (i-k+1,i-k+1, d)  
                    output_flatten = output_shape[0] * output_shape[1] * output_shape[2]
                    #if output_shape == layer_to.input_shape:
                    actions.append(Add_Res_Conv_Layer(pair))                
                elif type(layer_to) == Layer:
                    o = layer_to.input_size
                    i = layer_from.output_shape[0]
                    d = np.clip(math.floor(o ** ( 1 / 3)), 1, 10)
                    k = np.clip(i + 1 - math.ceil(o ** ( 1 / 3)), 1, i)
                    output_shape = (i-k+1,i-k+1, d)  
                    output_flatten = output_shape[0] * output_shape[1] * output_shape[2]
                    #if output_flatten == layer_to.input_size:
                    actions.append(Add_Res_Conv_Layer(pair))
        return actions
    
    def __str__(self):
        return " ( Add Res Conv Layer Action: " + str(self.params) + " ) "


import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import numpy as np
import unittest
from testSuite import mode

class TestingStorage(unittest.TestCase):

    def setUp(self):
        global mode
        if mode == 'cpu':
            gnn.switch_to_cpu()
        elif mode == 'gpu':
            gnn.switch_to_gpu()

    def test_base_save_load(self):
        M = gnn.structure.Model(3, 3, 1, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1)
        x = np.random.rand(3, 3)
        y = np.random.randint(2, size=(50,))
        output1 = M.forward_prop(x)
        gnn.Storage.saveModel(M, "here.json")
        M_loaded = gnn.Storage.loadModel("here.json")
        output2 = M_loaded.forward_prop(x)
        self.assertAlmostEqual(np.sum(output1 - output2), 0)

    def test_train_save_load(self):
        x_train = np.random.randint(0, 256, (5, 32, 32, 3), dtype=np.uint8)
        y_train = np.random.randint(0, 2, (5,), dtype=np.uint8)
        labels = {'frog': 0, 'cat': 1}
        for g in [2,3,10]:
            M = gnn.trainer.train(
            x_train = x_train, 
            y_train = y_train, 
            x_test= x_train,
            y_test= y_train,
            labels=labels,
            input_paths = 1,
            path = "ignore/", 
            model_name = "GNN_model", #"GNN_model", 
            epochs = 2, 
            generations = g,
            simulation_scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.PROGRESS_CHECK, simulation_time = 6, simulation_epochs = 2),
            simulation_alg = gnn.montecarlo_alg,
            sim_set_generator = gnn.simulation.create_simulation_set_SAMLE,
            lr_scheduler = gnn.LearningRateScheduler(gnn.LearningRateScheduler.PROGRESIVE_PARABOIDAL, 0.05, 0.8),
            input_size = 32 * 32, 
            hidden_size = 300, 
            output_size = 2, 
            input_shape = (32, 32, 3), 
            kernel_size = 3, 
            deepth = 2, 
            batch_size = 5 * 128, 
            simulation_set_size = 5,
            simulation_score = gnn.Simulation_score())
            output1 = M.forward_prop(x_train[:1])
            gnn.Storage.saveModel(M, "tmp.json")
            M_loaded = gnn.Storage.loadModel("tmp.json")
            output2 = M_loaded.forward_prop(x_train[:1])
            self.assertAlmostEqual(np.sum(output1 - output2), 0)
            

if __name__ == '__main__':
    unittest.main()



from .action import *
from .painter import *
from .structure import *
from .trainer import *
from .helpers import *
from .config import *
from .quaziIdentity import *

IS_CUPY = False

def switch_to_gpu():
   # print(" parent: switch_to_gpu")
    global np, IS_CUPY
    try:
        helpers.switch_to_gpu()
        structure.switch_to_gpu()
        import cupy as np
        IS_CUPY = False
        print(f"Cupy library loaded, GPU enabled")
    except Exception as e:
        print(f"Unexpected error occurred: {e} while loading cupy library, switching to CPU")
        switch_to_cpu()  # Fallback to CPU

def switch_to_cpu():
#    print(" parent: switch_to_cpu")
    global IS_CUPY, np
    helpers.switch_to_cpu()
    structure.switch_to_cpu()
    import numpy as np
    IS_CUPY = True
    print("Numpy library loaded, CPU enabled.")

switch_to_cpu()


import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import numpy as np
import unittest
import random
from testSuite import mode

shape = 20
epochs = 1

class TestingTrain(unittest.TestCase):

    def test_actions(self):
        M = gnn.structure.Model(shape,shape,2, gnn.structure.Loss.multiclass_cross_entropy, gnn.structure.Activations.Sigmoid, 1)
        x = np.random.rand(shape, shape)
        y = np.random.randint(2, size=(shape,))
        lr_scheduler = gnn.structure.LearningRateScheduler(gnn.structure.LearningRateScheduler.PROGRESIVE, 0.03, 0.8)
        for i in range(0,10):
            all_actions = gnn.action.Action.generate_all_actions(M)
            new_action = random.choice(all_actions)        
            new_action.execute(M)
        M.gradient_descent(x, y, 1, lr_scheduler, True)
        simulation_score = gnn.Simulation_score(0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5)
        score = simulation_score.scoreFun(M, 10, x, y)
        print("score: ", score)
        

if __name__ == '__main__':
    unittest.main()



import unittest
import argparse
import sys

# Define the test file names
test_files = [
    "testHistory.py",
    "testLayer.py",
    "testLRscheduler.py",
    "testModel.py",
    "testSimulationScheduler.py",
    "testSimulationScore.py",
    "testStorage.py",
    "unittests_autogenerated.py",
    "unittests_basic.py",
    "testTrain.py",
    "testOptimizers.py"
]

# Argument parser for CPU/GPU switch
parser = argparse.ArgumentParser(description="Run test suite with CPU/GPU mode.")
parser.add_argument('--mode', choices=['cpu', 'gpu'], default='cpu', help="Select computation mode: 'cpu' or 'gpu'")
args = parser.parse_args()
mode = args.mode

# Create a test loader
loader = unittest.TestLoader()

# Create an empty test suite
suite = unittest.TestSuite()

# Add tests from each test file to the test suite
for test_file in test_files:
    # Discover tests in the test file
    discovered_tests = loader.discover('./tests/', pattern=test_file)
    suite.addTests(discovered_tests)

# Create a test runner that will output the results to the console
runner = unittest.TextTestRunner()

# Run the test suite and exit with appropriate exit code
if __name__ == '__main__':
    result = runner.run(suite)
    if result.wasSuccessful():
        sys.exit(0)  # Exit with code 0 if all tests passed successfully
    else:
        sys.exit(1)  # Exit with code 1 if there were failures or errors in tests



from setuptools import setup, find_packages
import pkg_resources
import codecs
import os
# 1. change version
# 2. python setup.py sdist bdist_wheel
# 3. twine upload dist/*
here = os.path.abspath(os.path.dirname(__file__))

with open('LICENSE.txt') as f:
    license = f.read()
with open('README.md') as f:
    readme = f.read()
VERSION = '0.3.0'
DESCRIPTION = 'growingnn is a cutting-edge Python package that introduces a dynamic neural network architecture learning algorithm. This innovative approach allows the neural network to adapt its structure during training, optimizing both weights and architecture. Leveraging a Stochastic Gradient Descent-based optimizer and guided Monte Carlo tree search, the package provides a powerful tool for enhancing model performance ICCS 2025 update.'


# Setting up
setup(
    name="growingnn",
    version=VERSION,
    author="Szymon Świderski",
    author_email="<szymonswiderski.ai@gmail.com>",
    description=DESCRIPTION,
    license=license,
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires = ["numpy", "matplotlib", "opencv-python", "scipy", "imgaug", "scikit-learn", "pyvis"],
    keywords=['python', 'neural network', 'growing neural network', 'growing'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"])



import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import unittest
from testSuite import mode
from unittest.mock import MagicMock

class TestSimulationScheduler(unittest.TestCase):

    def setUp(self):
        global mode
        if mode == 'cpu':
            gnn.switch_to_cpu()
        elif mode == 'gpu':
            gnn.switch_to_gpu()
        self.hist_detail_mock = MagicMock()
        self.hist_detail_mock.Y = {
            'iteration_acc_train': [],
            'accuracy': []
        }

    def test_constant_mode_simulation(self):
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.CONSTANT, 100, 10)
        result = scheduler.can_simulate(1, self.hist_detail_mock)
        self.assertTrue(result)

    def test_progress_check_simulation_no_progress(self):
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.PROGRESS_CHECK, 100, 10)
        self.hist_detail_mock.Y['iteration_acc_train'] = [0.9, 0.9]
        result = scheduler.can_simulate(1, self.hist_detail_mock)
        self.assertTrue(result)

    def test_progress_check_simulation_with_progress(self):
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.PROGRESS_CHECK, 100, 10)
        self.hist_detail_mock.Y['iteration_acc_train'] = [0.8, 0.9]
        result = scheduler.can_simulate(1, self.hist_detail_mock)
        self.assertFalse(result)

    def test_overfit_check_simulation_overfit_detected(self):
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.OVERFIT_CHECK, 100, 10)
        self.hist_detail_mock.learning_capable.return_value = True
        result = scheduler.can_simulate(1, self.hist_detail_mock)
        self.assertTrue(result)

    def test_overfit_check_simulation_no_overfit(self):
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.OVERFIT_CHECK, 100, 10)
        self.hist_detail_mock.learning_capable.return_value = False
        result = scheduler.can_simulate(1, self.hist_detail_mock)
        self.assertFalse(result)

    def test_get_mode_label(self):
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.CONSTANT, 100, 10)
        self.assertEqual(scheduler.get_mode_label(), "simconstant")
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.PROGRESS_CHECK, 100, 10)
        self.assertEqual(scheduler.get_mode_label(), "simprogres")
        scheduler = gnn.SimulationScheduler(gnn.SimulationScheduler.OVERFIT_CHECK, 100, 10)
        self.assertEqual(scheduler.get_mode_label(), "simoverfit")

if __name__ == '__main__':
    unittest.main()



from pyvis.network import Network
from .helpers import *
import numpy as np

DEBUG = False
def draw(model, path = 'mygraph1.html'):
    net = Network(directed =True)
    edges = []
    if DEBUG: print("1 + ")
    for input_layer in model.input_layers:
        edges += draw_rec(model, input_layer.id, None, net, [])
    if DEBUG: print("2 + ")
    edges = delete_repetitions(edges)
    if DEBUG: print("3 + ")
    for edge in edges:
        net.add_edge(edge[0],edge[1])
    if DEBUG: print("4 + ")
    website = net.generate_html(path, False, False)
    with open(path, 'w', encoding='utf-8') as file:
        file.write(website)

def draw_rec(model, layer_id, parent_id, net, edges, deepth = 0):
    deepth += 1
    layer = model.get_layer(layer_id)
        
    # Default node color (low E)
    node_color = "#a6dced"  # Light blue in hexadecimal

    # Check if the layer has the matrix E and calculate its mean
    if hasattr(layer, 'E') and layer.E is not None:
        E_min = 0 
        E_max = np.prod(layer.E.shape)
        E_sum = np.sum(np.abs(layer.E))
        
        # Normalize E_mean to a range of 0 (light blue) to 1 (red)
        normalized_E = min(max((E_sum - E_min) / (E_max - E_min), 0), 1)
        
        # Light blue RGB values
        base_red = 166
        base_green = 220
        base_blue = 237

        # Calculate interpolated color
        red = int(base_red + (255 - base_red) * normalized_E)
        green = int(base_green + (0 - base_green) * normalized_E)
        blue = int(base_blue + (0 - base_blue) * normalized_E)
        
        # Construct the hexadecimal color
        node_color = f"#{red:02X}{green:02X}{blue:02X}"


    net.add_node(layer_id, layer.get_paint_label(), color=node_color)
    if parent_id!=None:
        edges.append([parent_id, layer_id])
    for child_layer_id in layer.output_layers_ids:
        edges += draw_rec(model, child_layer_id, layer_id, net,[], deepth)
    return edges





from ..helpers import *
from ..action import *
from ..structure import *
#from sklearn.decomposition import PCA

# Simulation set

def create_simulation_set_SAMLE(x,y, amount = 20):
    #set_size_ratio = float(amount) / float(max(np.shape(y)))
    x_sim, y_sim = protected_sampling(x, y, amount)
    #_, x_sim, _, y_sim = train_test_split(x, y, test_size=set_size_ratio)
    return x_sim, y_sim

# def create_simulation_set_PCA(x,y, amount = 10):
#     #print("1 x shape: ", x.shape)
#     #print("22: x: ", x.shape[0], " y: ", y.shape[0])
#     reverse_axes = False
#     if x.shape[0] != y.shape[0]:
#         x = np.swapaxes(x, 0, 1)
#         reverse_axes = True
#         #print("2 x shape: ", x.shape)
#     if y.shape[0] < 100: return create_simulation_set_SAMLE(x,y,amount)
#     if x.shape[0] != y.shape[0]: print("error wrong dimensions !")
#     pca = PCA(amount)
#     classes = {}
#     x_result = []
#     y_result = []
#     for i in range(x.shape[0]):
#         if not y[i] in classes.keys():
#             classes[y[i]] = []
#         classes[y[i]].append(x[i])
#     for c in classes:
#         x_per_class = np.array(classes[c])
#         needed_shape = (amount,) + x_per_class.shape[1:]
#         n = np.prod(x_per_class.shape[1:]) # obliczamy długość wektora spłaszczonego
#         x_per_class_flat = x_per_class.reshape(x_per_class.shape[0], n)
#         x_per_class_flat = x_per_class_flat.T
#         sim_x_per_class_flat = pca.fit_transform(x_per_class_flat)
#         sim_x_per_class = sim_x_per_class_flat.reshape(needed_shape)
#         x_result.append(sim_x_per_class)
#         y_result.append(np.ones((amount,)) * int(c))
#     x_result = np.array(x_result)
#     y_result = np.array(y_result).astype(int)
#     x_result = np.concatenate(x_result, axis=0)
#     y_result = np.concatenate(y_result, axis=0)
#     print("3 x shape: ", x.shape)
#     if reverse_axes:
#         x = np.swapaxes(x, 0, 1)
#         print("4 x shape: ", x.shape)
#         print("x_result: ", x_result.shape)
#         print("y_result: ", y_result.shape)
#         x_result = np.swapaxes(x_result, 0, 1)
#     print("x_result: ", x_result.shape)
#     print("y_result: ", y_result.shape)
#     return x_result, y_result


import time
from ...structure import *

def scoreTime(M, epochs, X_train, Y_train):
    # more time smaller score
    start_time = time.time()
    _, _ = M.gradient_descent(X_train, Y_train, 1, LearningRateScheduler(LearningRateScheduler.CONSTANT, 0.1) , True)
    end_time = time.time()
    time_difference_in_seconds = end_time - start_time
    return (1/(time_difference_in_seconds + 1))


def scoreCountWeights(M, epochs, X_train, Y_train):
    # more time smaller score
    counter = 0
    for layer_id in M.hidden_layers:
        layer = M.get_layer(layer_id)
        if type(layer) == Conv:
            counter += int(layer.depth) * int(layer.input_depth) * int(layer.kernel_size) * int(layer.kernel_size)
        elif type(layer) == Layer:
            counter += layer.input_size * layer.neurons
    return (1/(counter + 1))



import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import unittest
from testSuite import mode

class TestLearningRateScheduler(unittest.TestCase):

    def setUp(self):
        global mode
        if mode == 'cpu':
            gnn.switch_to_cpu()
        elif mode == 'gpu':
            gnn.switch_to_gpu()

    def test_constant_scheduler(self):
        scheduler = gnn.LearningRateScheduler(mode=gnn.LearningRateScheduler.CONSTANT, alpha=0.1)
        alpha = scheduler.alpha_scheduler(i=10, iterations=100)
        self.assertEqual(alpha, 0.1)
        
    def test_progressive_scheduler(self):
        scheduler = gnn.LearningRateScheduler(mode=gnn.LearningRateScheduler.PROGRESIVE, alpha=0.1, steepness=0.2)
        alpha = scheduler.alpha_scheduler(i=10, iterations=100)
        expected = 0.1 * (11 / 22)
        self.assertAlmostEqual(alpha, expected) 
        
    def test_progressive_scheduler_min_max_PROGRESIVE(self):
        alpha_max = 0.1
        scheduler = gnn.LearningRateScheduler(mode=gnn.LearningRateScheduler.PROGRESIVE, alpha=alpha_max, steepness=0.2)
        values = []
        for i in range(0, 100):
            values.append(scheduler.alpha_scheduler(i, 100))
        self.assertTrue(max(values) == alpha_max) 
        self.assertTrue(min(values) <= alpha_max * 0.1) 
        self.assertTrue(min(values) >= 0) 
        
    def test_progressive_scheduler_min_max_PROGRESIVE_PARABOIDAL(self):
        alpha_max = 0.1
        scheduler = gnn.LearningRateScheduler(mode=gnn.LearningRateScheduler.PROGRESIVE_PARABOIDAL, alpha=alpha_max, steepness=0.2)
        values = []
        for i in range(0, 100):
            values.append(scheduler.alpha_scheduler(i, 100))
        print("values: ", values)
        print("min(values): ", min(values))
        print("max(values): ", max(values))
        self.assertTrue(max(values) == alpha_max) 
        self.assertTrue(min(values) <= alpha_max * 0.1) 
        self.assertTrue(min(values) >= 0) 



if __name__ == '__main__':
    unittest.main()



from ...structure import *
from .scoreByLearning import *
from .scoreEfficiency import *
from .scoreGraphStructure import *

class Simulation_score:
    #ACCURACY = 0
    #LOSS = 1
    # def __init__(self, mode = 0):
    #     self.mode = mode

    ScoreFunctions = {
        'weight_acc': scoreAcc,
        'weight_loss': scoreLoss,
        'weight_time': scoreTime,
        'weight_countW': scoreCountWeights,
        'weight_diameter': scoreDiameter,
        'weight_radius': scoreRadius,
        'weight_eccentricity': scoreEccentricity,
        'weight_chromaticN': scoreChromaticNumber,
        'weight_machingN': scoreMatchingNumber,
        'weight_independenceN': scoreIndependenceNumber,
    }
    def __init__(self, 
                 weight_acc =  1.0, 
                 weight_loss = 0.0,
                 weight_time = 0.0,
                 weight_countW = 0.0,
                 weight_diameter = 0.0,
                 weight_radius = 0.0,
                 weight_eccentricity = 0.0,
                 weight_chromaticN = 0.0,
                 weight_machingN = 0.0,
                 weight_independenceN = 0.0):
        self.weights = {}
        self.weights['weight_acc'] = weight_acc
        self.weights['weight_loss'] = weight_loss
        self.weights['weight_time'] = weight_time
        self.weights['weight_countW'] = weight_countW
        self.weights['weight_diameter'] = weight_diameter
        self.weights['weight_radius'] = weight_radius
        self.weights['weight_eccentricity'] = weight_eccentricity
        self.weights['weight_chromaticN'] = weight_chromaticN
        self.weights['weight_machingN'] = weight_machingN
        self.weights['weight_independenceN'] = weight_independenceN

        #     self.mode = mode
    # def new_max_loss(self, global_history):
    #     self.max_loss = numpy.max(get_list_as_numpy_array(global_history.Y['loss']))
        
    # def grade(self, acc, history):
    #     if self.mode == Simulation_score.ACCURACY:
    #         return acc
    #     else:
    #         return max(1.e-17, self.max_loss - history.get_last('loss'))
            
    def weightSum(self):
        sum = 0.0
        for key in self.weights.keys():
            sum += self.weights[key]
        return sum

    def scoreFun(self, M, epochs, X_train, Y_train):
        score = 0.0
        for score_weight in Simulation_score.ScoreFunctions.keys():
            if self.weights[score_weight] > 0.0:
                score  += self.weights[score_weight] * Simulation_score.ScoreFunctions[score_weight](M, epochs, X_train, Y_train)
        return score / self.weightSum()